import logging
from typing import Dict, List

from openfisca_core.model_api import MONTH, YEAR
from openfisca_core.simulations.simulation_builder import SimulationBuilder


SimulationBuilder.id_variable_by_entity_key = None
SimulationBuilder.role_variable_by_entity_key = None
SimulationBuilder.used_as_input_variables = None
SimulationBuilder.used_as_input_variables_by_entity = None


log = logging.getLogger(__name__)


# Helpers

def diagnose_variable_mismatch(used_as_input_variables, input_data_frame):
    """Diagnose variables mismatch.

    Args:
      used_as_input_variables(lsit): List of variable to test presence
      input_data_frame: DataFrame in which to test variables presence

    """
    variables_mismatch = set(used_as_input_variables).difference(set(input_data_frame.columns)) if used_as_input_variables else None
    if variables_mismatch:
        log.info(
            'The following variables are used as input variables are not present in the input data frame: \n {}'.format(
                sorted(variables_mismatch)))
    if variables_mismatch:
        log.debug('The following variables are used as input variables: \n {}'.format(
            sorted(used_as_input_variables)))
        log.debug('The input_data_frame contains the following variables: \n {}'.format(
            sorted(list(input_data_frame.columns))))


# SimulationBuilder monkey-patched methods

def _set_id_variable_by_entity_key(builder) -> Dict[str, str]:
    """Identify and sets the correct ids for the different entities."""
    if builder.id_variable_by_entity_key is None:
        log.debug("Use default id_variable names")
        builder.id_variable_by_entity_key = dict(
            (entity.key, entity.key + '_id') for entity in builder.tax_benefit_system.entities)

    return builder.id_variable_by_entity_key


def _set_role_variable_by_entity_key(builder) -> Dict[str, str]:
    """Identify and sets the correct roles for the different entities."""
    if builder.role_variable_by_entity_key is None:
        builder.role_variable_by_entity_key = dict(
            (entity.key, entity.key + '_role_index') for entity in builder.tax_benefit_system.entities)

    return builder.role_variable_by_entity_key


def _set_used_as_input_variables_by_entity(builder) -> Dict[str, List[str]]:
    """Identify and sets the correct input variables for the different entities."""
    if builder.used_as_input_variables_by_entity is not None:
        return

    tax_benefit_system = builder.tax_benefit_system

    assert set(builder.used_as_input_variables) <= set(tax_benefit_system.variables.keys()), \
        "Some variables used as input variables are not part of the tax benefit system:\n {}".format(
            set(builder.used_as_input_variables).difference(set(tax_benefit_system.variables.keys()))
            )

    builder.used_as_input_variables_by_entity = dict()

    for entity in tax_benefit_system.entities:
        builder.used_as_input_variables_by_entity[entity.key] = [
            variable
            for variable in builder.used_as_input_variables
            if tax_benefit_system.get_variable(variable).entity.key == entity.key
            ]

    return builder.used_as_input_variables_by_entity


def filter_input_variables(builder, input_data_frame, tax_benefit_system):
    """Filter the input data frame from variables that won't be used or are set to be computed.

    Args:
        input_data_frame: Input dataframe (Default value = None)

    Returns:
        pd.DataFrame: filtered dataframe

    """
    assert input_data_frame is not None
    id_variable_by_entity_key = builder.id_variable_by_entity_key
    role_variable_by_entity_key = builder.role_variable_by_entity_key
    used_as_input_variables = builder.used_as_input_variables

    variables = tax_benefit_system.variables

    id_variables = [
        id_variable_by_entity_key[_entity.key] for _entity in tax_benefit_system.group_entities]
    role_variables = [
        role_variable_by_entity_key[_entity.key] for _entity in tax_benefit_system.group_entities]

    log.debug('Variable used_as_input_variables in filter: \n {}'.format(used_as_input_variables))

    unknown_columns = []
    for column_name in input_data_frame:
        if column_name in id_variables + role_variables:
            continue
        if column_name not in variables:
            unknown_columns.append(column_name)

    input_data_frame.drop(unknown_columns, axis = 1, inplace = True)

    if unknown_columns:
        log.debug('The following unknown columns {}, are dropped from input table'.format(
            sorted(unknown_columns)))

    used_columns = []
    dropped_columns = []
    for column_name in input_data_frame:
        if column_name in id_variables + role_variables:
            continue
        variable = variables[column_name]
        # Keeping the calculated variables that are initialized by the input data
        if variable.formulas:
            if column_name in used_as_input_variables:
                used_columns.append(column_name)
                continue

            dropped_columns.append(column_name)

    input_data_frame.drop(dropped_columns, axis = 1, inplace = True)

    if used_columns:
        log.debug(
            'These columns are not dropped because present in used_as_input_variables:\n {}'.format(
                sorted(used_columns)))
    if dropped_columns:
        log.debug(
            'These columns in survey are set to be calculated, we drop them from the input table:\n {}'.format(
                sorted(dropped_columns)))

    log.info('Keeping the following variables in the input_data_frame:\n {}'.format(
        sorted(list(input_data_frame.columns))))
    return input_data_frame


def init_all_entities(builder, input_data_frame, period = None):
    assert period is not None
    log.info('Initialasing simulation using input_data_frame for period {}'.format(period))
    builder._set_id_variable_by_entity_key()
    builder._set_role_variable_by_entity_key()

    if period.unit == YEAR:  # 1. year
        simulation = builder.init_simulation_with_data_frame(
            input_data_frame = input_data_frame,
            period = period,
            )
    elif period.unit == MONTH and period.size == 3:  # 2. quarter
        for offset in range(period.size):
            period_item = period.first_month.offset(offset, MONTH)
            simulation = builder.init_simulation_with_data_frame(
                input_data_frame = input_data_frame,
                period = period_item,
                )
    elif period.unit == MONTH and period.size == 1:  # 3. months
        simulation = builder.init_simulation_with_data_frame(
            input_data_frame = input_data_frame,
            period = period,
            )
    else:
        raise ValueError("Invalid period {}".format(period))

    simulation.id_variable_by_entity_key = builder.id_variable_by_entity_key
    return simulation


def init_entity_structure(builder, entity, input_data_frame):
    """Initialize sthe simulation with tax_benefit_system entities and input_data_frame.

    Args:
        tax_benefit_system(TaxBenfitSystem): The TaxBenefitSystem to get the structure from
        entity(Entity): The entity to initialize structure
        input_data_frame(pd.DataFrame): The input
        builder(Builder): The builder

    """
    tax_benefit_system = builder.tax_benefit_system
    builder._set_id_variable_by_entity_key()
    builder._set_role_variable_by_entity_key()
    builder._set_used_as_input_variables_by_entity()

    input_data_frame = builder.filter_input_variables(input_data_frame, tax_benefit_system)

    id_variables = [
        builder.id_variable_by_entity_key[_entity.key] for _entity in tax_benefit_system.group_entities]
    role_variables = [
        builder.role_variable_by_entity_key[_entity.key] for _entity in tax_benefit_system.group_entities]

    if entity.is_person:
        for id_variable in id_variables + role_variables:
            assert id_variable in input_data_frame.columns, \
                "Variable {} is not present in input dataframe".format(id_variable)

    ids = range(len(input_data_frame))
    if entity.is_person:
        builder.declare_person_entity(entity.key, ids)
        for group_entity in tax_benefit_system.group_entities:
            _key = group_entity.key
            _id_variable = builder.id_variable_by_entity_key[_key]
            _role_variable = builder.role_variable_by_entity_key[_key]
            group_population = builder.declare_entity(_key, input_data_frame[_id_variable].drop_duplicates().sort_values().values)
            builder.join_with_persons(
                group_population,
                input_data_frame[_id_variable].astype('int').values,
                input_data_frame[_role_variable].astype('int').values,
                )


def init_simulation_with_data_frame(builder, input_data_frame, period):
    """Initialize the simulation period with current input_data_frame for an entity if specified."""
    used_as_input_variables = builder.used_as_input_variables
    id_variable_by_entity_key = builder.id_variable_by_entity_key
    role_variable_by_entity_key = builder.role_variable_by_entity_key
    tax_benefit_system = builder.tax_benefit_system
    assert tax_benefit_system is not None

    diagnose_variable_mismatch(used_as_input_variables, input_data_frame)

    id_variables = [
        id_variable_by_entity_key[_entity.key] for _entity in tax_benefit_system.group_entities]
    role_variables = [
        role_variable_by_entity_key[_entity.key] for _entity in tax_benefit_system.group_entities]

    for id_variable in id_variables + role_variables:
        assert id_variable in input_data_frame.columns, \
            "Variable {} is not present in input dataframe".format(id_variable)

    input_data_frame = builder.filter_input_variables(input_data_frame, tax_benefit_system)

    index_by_entity_key = dict()

    for entity in tax_benefit_system.entities:
        builder.init_entity_structure(entity, input_data_frame)

        if entity.is_person:
            continue

        else:
            index_by_entity_key[entity.key] = input_data_frame.loc[
                input_data_frame[role_variable_by_entity_key[entity.key]] == 0,
                id_variable_by_entity_key[entity.key]
                ].sort_values().index

    for column_name, column_serie in input_data_frame.items():
        if role_variable_by_entity_key is not None:
            if column_name in role_variable_by_entity_key.values():
                continue

        if id_variable_by_entity_key is not None:
            if column_name in id_variable_by_entity_key.values():
                continue

        simulation = builder.build(tax_benefit_system)
        entity = tax_benefit_system.variables[column_name].entity
        if entity.is_person:
            simulation.init_variable_in_entity(entity.key, column_name, column_serie, period)
        else:
            simulation.init_variable_in_entity(entity.key, column_name, column_serie[index_by_entity_key[entity.key]], period)

    assert builder.id_variable_by_entity_key is not None
    simulation.id_variable_by_entity_key = builder.id_variable_by_entity_key
    return simulation


SimulationBuilder._set_id_variable_by_entity_key = _set_id_variable_by_entity_key
SimulationBuilder._set_role_variable_by_entity_key = _set_role_variable_by_entity_key
SimulationBuilder._set_used_as_input_variables_by_entity = _set_used_as_input_variables_by_entity
SimulationBuilder.filter_input_variables = filter_input_variables
SimulationBuilder.init_all_entities = init_all_entities
SimulationBuilder.init_entity_structure = init_entity_structure
SimulationBuilder.init_simulation_with_data_frame = init_simulation_with_data_frame



import os
import pandas as pd
import pytest


from openfisca_survey_manager import openfisca_survey_manager_location
from openfisca_survey_manager.survey_collections import SurveyCollection
from openfisca_survey_manager.scripts.build_collection import add_survey_to_collection
from openfisca_survey_manager.input_dataframe_generator import set_table_in_survey


@pytest.mark.order(after="test_write_parquet.py::TestWriteParquet::test_write_parquet_one_file_per_entity")
def test_add_survey_to_collection_parquet():
    name = 'fake'
    survey_name = 'test_parquet'
    data_dir = os.path.join(
        openfisca_survey_manager_location,
        'openfisca_survey_manager',
        'tests',
        'data_files',
        )
    survey_collection = SurveyCollection(name = name)
    saved_fake_survey_file_path = os.path.join(data_dir, 'test.parquet')
    add_survey_to_collection(
        survey_name = survey_name,
        survey_collection = survey_collection,
        parquet_files = [saved_fake_survey_file_path],
        )
    ordered_dict = survey_collection.to_json()
    assert survey_name in list(ordered_dict['surveys'].keys())


@pytest.mark.order(after="test_write_parquet.py::TestWriteParquet::test_write_parquet_one_file_per_entity")
def test_set_table_in_survey_parquet():
    data_dir = os.path.join(
        openfisca_survey_manager_location,
        'openfisca_survey_manager/tests/data_files',
        )
    filepath = os.path.join(data_dir, 'test_parquet_collection', 'household.parquet')
    input_dataframe = pd.read_parquet(filepath)
    survey_name = 'test_parquet'
    collection = "fake"
    set_table_in_survey(input_dataframe, entity="foyer", period="2023", collection = collection, survey_name = survey_name, config_files_directory=data_dir)

    # Read survey
    survey_collection = SurveyCollection.load(config_files_directory = data_dir, collection=collection)
    survey = survey_collection.get_survey(survey_name)
    table = survey.get_values(
        table="foyer_2023", ignorecase=True
        )
    assert len(table) == 4
    assert (table.columns == ["household_id", "rent", "household_weight", "accommodation_size"]).all()
    assert table.household_weight.sum() == 2950


def test_add_survey_to_collection():
    name = 'fake'
    survey_name = 'fake_survey'
    data_dir = os.path.join(
        openfisca_survey_manager_location,
        'openfisca_survey_manager',
        'tests',
        'data_files',
        )
    survey_collection = SurveyCollection(name = name)
    saved_fake_survey_file_path = os.path.join(data_dir, 'help.sas7bdat')
    add_survey_to_collection(
        survey_name = survey_name,
        survey_collection = survey_collection,
        sas_files = [saved_fake_survey_file_path],
        stata_files = []
        )
    ordered_dict = survey_collection.to_json()
    assert survey_name in list(ordered_dict['surveys'].keys())


@pytest.mark.order(after="test_surveys.py::test_survey")
def test_set_table_in_survey_first_year():
    data_dir = os.path.join(
        openfisca_survey_manager_location,
        'openfisca_survey_manager/tests/data_files',
        )
    input_dataframe = pd.DataFrame({"rfr": [1_000, 2_000, 100_000]})
    survey_name = 'test_set_table_in_survey_2020'
    collection = "fake"
    set_table_in_survey(input_dataframe, entity="foyer", period="2020", collection = collection, survey_name = survey_name, config_files_directory=data_dir)

    # Read survey
    survey_collection = SurveyCollection.load(config_files_directory = data_dir, collection=collection)
    survey = survey_collection.get_survey(survey_name)
    table = survey.get_values(
        table="foyer_2020", ignorecase=True
        )
    assert len(table) == 3
    assert table.columns == ["rfr"]
    assert table.rfr.sum() == 103000


@pytest.mark.order(after="test_set_table_in_survey_first_year")
def test_set_table_in_survey_second_year():
    data_dir = os.path.join(
        openfisca_survey_manager_location,
        'openfisca_survey_manager/tests/data_files',
        )
    input_dataframe = pd.DataFrame({"rfr": [1_021, 2_021, 100_021]})
    survey_name = 'test_set_table_in_survey_2021'
    collection = "fake"
    set_table_in_survey(input_dataframe, entity="foyer", period="2021", collection = collection, survey_name = survey_name, config_files_directory=data_dir)

    # Read first survey
    survey_name = 'test_set_table_in_survey_2020'
    survey_collection = SurveyCollection.load(config_files_directory = data_dir, collection=collection)
    survey = survey_collection.get_survey(survey_name)
    table = survey.get_values(
        table="foyer_2020", ignorecase=True
        )
    assert len(table) == 3
    assert table.columns == ["rfr"]
    assert table.rfr.sum() == 103000

    # Read second survey
    survey_name = 'test_set_table_in_survey_2021'
    survey_collection = SurveyCollection.load(config_files_directory = data_dir, collection=collection)
    survey = survey_collection.get_survey(survey_name)
    table = survey.get_values(
        table="foyer_2021", ignorecase=True
        )
    assert len(table) == 3
    assert table.columns == ["rfr"]
    assert table.rfr.sum() == 103063



"""Abstract survey scenario definition."""

import logging
import numpy as np
import pandas as pd

from typing import Optional, Union
from openfisca_core.types import Array, Period

from openfisca_survey_manager.scenarios.abstract_scenario import AbstractSurveyScenario
from openfisca_survey_manager.simulations import Simulation


log = logging.getLogger(__name__)


class ReformScenario(AbstractSurveyScenario):
    """Reform survey scenario."""

    def _get_simulation(self, use_baseline: bool = False):
        """
        Get relevant simulation

        Args:
            use_baseline (bool, optional): Whether to get baseline or reform simulation. Defaults to False.
        """

        if len(self.simulations) == 1:
            return list(self.simulations.values())[0]

        simulation_name = (
            "baseline"
            if use_baseline
            else "reform"
            )
        simulation = self.simulations[simulation_name]
        assert simulation is not None, f"{simulation_name} does not exist"
        return simulation

    def build_input_data(self, **kwargs):
        """Build input data."""
        NotImplementedError

    def calculate_series(self, variable, period = None, use_baseline = False):
        """Compute variable values for period and baseline or reform tax benefit and system.

        Args:
          variable(str, optional): Variable to compute
          period(Period, optional): Period, defaults to None
          use_baseline(bool, optional): Use baseline tax and benefit system, defaults to False

        Returns:
          pandas.Series: Variable values

        """
        return pd.Series(
            data = self.calculate_variable(variable, period, use_baseline),
            name = variable,
            )

    def calculate_variable(self, variable, period = None, use_baseline = False):
        """Compute variable values for period and baseline or reform tax benefit and system.

        Args:
          variable(str, optional): Variable to compute
          period(Period, optional): Period, defaults to None
          use_baseline(bool, optional): Use baseline tax and benefit system, defaults to False

        Returns:
          numpy.ndarray: Variable values

        """
        simulation = self._get_simulation(use_baseline)
        return simulation.adaptative_calculate_variable(variable, period = period)

    def compute_aggregate(self, variable: str = None, aggfunc: str = 'sum', filter_by: str = None,
            period: Optional[Union[int, str, Period]] = None, use_baseline: bool = False,
            difference: bool = False, missing_variable_default_value = np.nan, weighted: bool = True,
            alternative_weights: Optional[Union[str, int, float, Array]] = None):
        """Compute variable aggregate.

        Args:
            variable (str, optional): Variable to aggregate. Defaults to None.
            aggfunc (str, optional): Aggregation function. Defaults to 'sum'.
            filter_by (str, optional): Filter variable or expression to use. Defaults to None.
            period (Optional[Union[int, str, Period]], optional): Period. Defaults to None.
            use_baseline: Use baseline simulation. Defaults to False.
            missing_variable_default_value (optional): Value to use for missing values. Defaults to np.nan.
            weighted (bool, optional): Whether to weight the variable or not. Defaults to True.
            alternative_weights (Optional[Union[str, int, float, Array]], optional): Alternative weigh to use. Defaults to None.
            filtering_variable_by_entity (Dict, optional): Filtering variable by entity. Defaults to None.

        Returns:
            float: Aggregate
        """
        assert aggfunc in ['count', 'mean', 'sum', 'count_non_zero']
        assert period is not None
        assert not (difference and use_baseline), "Can't have difference and use_baseline both set to True"

        if difference:
            return (
                self.compute_aggregate(
                    variable = variable,
                    aggfunc = aggfunc,
                    filter_by = filter_by,
                    period = period,
                    use_baseline = False,
                    missing_variable_default_value = missing_variable_default_value,
                    weighted = weighted,
                    alternative_weights = alternative_weights,
                    )
                - self.compute_aggregate(
                    variable = variable,
                    aggfunc = aggfunc,
                    filter_by = filter_by,
                    period = period,
                    use_baseline = True,
                    missing_variable_default_value = missing_variable_default_value,
                    weighted = weighted,
                    alternative_weights = alternative_weights,
                    )
                )

        assert variable is not None
        simulation = self._get_simulation(use_baseline)
        return simulation.compute_aggregate(
            variable = variable,
            aggfunc = aggfunc,
            filter_by = filter_by,
            period = period,
            missing_variable_default_value = missing_variable_default_value,
            weighted = weighted,
            alternative_weights = alternative_weights,
            filtering_variable_by_entity = self.filtering_variable_by_entity,
            )

    def compute_quantiles(self, variable: str = None, nquantiles = None, period = None, use_baseline = False, filter_by = None,
            weighted = True, alternative_weights = None):

        assert variable is not None
        assert nquantiles is not None
        simulation = self._get_simulation(use_baseline)

        return simulation.compute_quantiles(
            variable = variable,
            period = period,
            nquantiles = nquantiles,
            filter_by = filter_by,
            weighted = weighted,
            alternative_weights = alternative_weights,
            )

    def compute_marginal_tax_rate(self, target_variable: str, period: Optional[Union[int, str, Period]], use_baseline: bool = False,
            value_for_zero_varying_variable: float = 0.0) -> Array:
        """
        Compute marginal a rate of a target (MTR) with respect to a varying variable.

        Args:
            target_variable (str): the variable which marginal tax rate is computed
            period (Optional[Union[int, str, Period]], optional): Period. Defaults to None.
            use_baseline: Use baseline simulation. Defaults to False.
            value_for_zero_varying_variable (float, optional): value of MTR when the varying variable is zero. Defaults to 0.

        Returns:
            numpy.array: Vector of marginal rates
        """
        if use_baseline:
            return super(ReformScenario, self).compute_marginal_tax_rate(
                target_variable = target_variable,
                period = period,
                simulation = "baseline",
                value_for_zero_varying_variable = value_for_zero_varying_variable,
                )
        else:
            return super(ReformScenario, self).compute_marginal_tax_rate(
                target_variable = target_variable,
                period = period,
                simulation = "reform",
                value_for_zero_varying_variable = value_for_zero_varying_variable,
                )

    def compute_pivot_table(self, aggfunc = 'mean', columns = None, difference = False, filter_by = None, index = None,
            period = None, use_baseline = False, use_baseline_for_columns = None, values = None,
            missing_variable_default_value = np.nan, concat_axis = None, weighted = True, alternative_weights = None):

        filtering_variable_by_entity = self.filtering_variable_by_entity

        return Simulation.compute_pivot_table(
            aggfunc = aggfunc,
            columns = columns,
            baseline_simulation = self._get_simulation(use_baseline = True),
            filter_by = filter_by,
            index = index,
            period = period,
            simulation = self._get_simulation(use_baseline),
            difference = difference,
            use_baseline_for_columns = use_baseline_for_columns,
            values = values,
            missing_variable_default_value = missing_variable_default_value,
            concat_axis = concat_axis,
            weighted = weighted,
            alternative_weights = alternative_weights,
            filtering_variable_by_entity = filtering_variable_by_entity
            )

    def compute_winners_loosers(self, variable = None,
            filter_by = None,
            period = None,
            absolute_minimal_detected_variation = 0,
            relative_minimal_detected_variation = .01,
            observations_threshold = None,
            weighted = True,
            alternative_weights = None):

        return super(ReformScenario, self).compute_winners_loosers(
            simulation = "reform",
            baseline_simulation = "baseline",
            variable = variable,
            filter_by = filter_by,
            period = period,
            absolute_minimal_detected_variation = absolute_minimal_detected_variation,
            relative_minimal_detected_variation = relative_minimal_detected_variation,
            observations_threshold = observations_threshold,
            weighted = weighted,
            alternative_weights = alternative_weights,
            )

    def create_data_frame_by_entity(self, variables = None, expressions = None, filter_by = None, index = False,
            period = None, use_baseline = False, merge = False):
        """Create dataframe(s) of computed variable for every entity (eventually merged in a unique dataframe).

        Args:
          variables(list, optional): Variable to compute, defaults to None
          expressions(str, optional): Expressions to compute, defaults to None
          filter_by(str, optional): Boolean variable or expression, defaults to None
          index(bool, optional): Index by entity id, defaults to False
          period(Period, optional): Period, defaults to None
          use_baseline(bool, optional): Use baseline tax and benefit system, defaults to False
          merge(bool, optional): Merge all the entities in one data frame, defaults to False

        Returns:
          dict or pandas.DataFrame: Dictionnary of dataframes by entities or dataframe with all the computed variables

        """
        simulation = self._get_simulation(use_baseline)
        return simulation.create_data_frame_by_entity(
            variables = variables,
            expressions = expressions,
            filter_by = filter_by,
            index = index,
            period = period,
            merge = merge,
            )



import shutil
import logging
import os
import pytest


from openfisca_core import periods
from openfisca_core.tools import assert_near


from openfisca_survey_manager import openfisca_survey_manager_location, default_config_files_directory
from openfisca_survey_manager.input_dataframe_generator import (
    make_input_dataframe_by_entity,
    random_data_generator,
    randomly_init_variable,
    )
from openfisca_survey_manager.scenarios.abstract_scenario import AbstractSurveyScenario
from openfisca_survey_manager.scenarios.reform_scenario import ReformScenario
from openfisca_survey_manager.tests import tax_benefit_system


log = logging.getLogger(__name__)


def create_randomly_initialized_survey_scenario(nb_persons = 10, nb_groups = 5, salary_max_value = 50000,
        rent_max_value = 1000, collection = "test_random_generator", use_marginal_tax_rate = False, reform = None):

    if collection is not None:
        return create_randomly_initialized_survey_scenario_from_table(
            nb_persons, nb_groups, salary_max_value, rent_max_value, collection, use_marginal_tax_rate, reform = reform)
    else:
        return create_randomly_initialized_survey_scenario_from_data_frame(
            nb_persons, nb_groups, salary_max_value, rent_max_value, use_marginal_tax_rate, reform = reform)


def create_randomly_initialized_survey_scenario_from_table(nb_persons, nb_groups, salary_max_value, rent_max_value, collection, use_marginal_tax_rate, reform = None):
    variable_generators_by_period = {
        periods.period('2017-01'): [
            {
                'variable': 'salary',
                'max_value': salary_max_value,
                },
            {
                'variable': 'rent',
                'max_value': rent_max_value,
                },
            {
                'variable': 'household_weight',
                'max_value': 100,
                },
            ],
        periods.period('2018-01'): [
            {
                'variable': 'salary',
                'max_value': salary_max_value,
                },
            ],
        }
    table_by_entity_by_period = random_data_generator(tax_benefit_system, nb_persons, nb_groups,
        variable_generators_by_period, collection)
    if reform is None:
        survey_scenario = AbstractSurveyScenario()
        survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
    else:
        survey_scenario = ReformScenario()
        survey_scenario.set_tax_benefit_systems(dict(
            reform = reform(tax_benefit_system),
            baseline = tax_benefit_system,
            ))

    survey_scenario.used_as_input_variables = ['salary', 'rent', 'housing_occupancy_status', 'household_weight']
    survey_scenario.period = 2017
    survey_scenario.collection = collection
    data = {
        'survey': 'input',
        'input_data_table_by_entity_by_period': table_by_entity_by_period,
        'config_files_directory': default_config_files_directory
        }
    if use_marginal_tax_rate:
        survey_scenario.varying_variable = 'salary'

    survey_scenario.weight_variable_by_entity = {
        "person": "person_weight",
        "household": "household_weight",
        }
    survey_scenario.init_from_data(data = data, use_marginal_tax_rate = use_marginal_tax_rate)
    return survey_scenario


def create_randomly_initialized_survey_scenario_from_data_frame(nb_persons, nb_groups, salary_max_value, rent_max_value, use_marginal_tax_rate = False, reform = None):
    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value)
    survey_scenario = AbstractSurveyScenario()
    weight_variable_by_entity = {
        "person": "person_weight",
        "household": "household_weight",
        }
    if reform is None:
        survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
    else:
        survey_scenario.set_tax_benefit_systems(dict(
            reform = reform(tax_benefit_system),
            baseline = tax_benefit_system,
            ))
    survey_scenario.period = 2017
    survey_scenario.used_as_input_variables = ['salary', 'rent', 'household_weight']
    period = periods.period('2017-01')

    data = {
        'input_data_frame_by_entity_by_period': {
            period: input_data_frame_by_entity
            },
        'config_files_directory': default_config_files_directory
        }
    survey_scenario.set_weight_variable_by_entity(weight_variable_by_entity)
    assert survey_scenario.weight_variable_by_entity == weight_variable_by_entity
    survey_scenario.init_from_data(data = data)
    for simulation_name, simulation in survey_scenario.simulations.items():
        assert simulation.weight_variable_by_entity == weight_variable_by_entity, f"{simulation_name} weight_variable_by_entity does not match {weight_variable_by_entity}"
        assert (survey_scenario.calculate_series("household_weight", period, simulation = simulation_name) != 0).all()
    return survey_scenario


def generate_input_input_dataframe_by_entity(nb_persons, nb_groups, salary_max_value, rent_max_value):
    input_dataframe_by_entity = make_input_dataframe_by_entity(tax_benefit_system, nb_persons, nb_groups)
    randomly_init_variable(
        tax_benefit_system,
        input_dataframe_by_entity,
        'salary',
        max_value = salary_max_value,
        condition = "household_role == 'first_parent'"
        )
    randomly_init_variable(
        tax_benefit_system,
        input_dataframe_by_entity,
        'rent',
        max_value = rent_max_value
        )
    randomly_init_variable(
        tax_benefit_system,
        input_dataframe_by_entity,
        'household_weight',
        max_value = 100,
        )
    return input_dataframe_by_entity


def test_input_dataframe_generator(nb_persons = 10, nb_groups = 5, salary_max_value = 50000,
        rent_max_value = 1000):
    input_dataframe_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value)
    assert (input_dataframe_by_entity['person']['household_role'] == "first_parent").sum() == 5
    assert (input_dataframe_by_entity['person'].loc[
        input_dataframe_by_entity['person']['household_role'] != "first_parent",
        'salary'
        ] == 0).all()

    assert (input_dataframe_by_entity['person'].loc[
        input_dataframe_by_entity['person']['household_role'] == "first_parent",
        'salary'
        ] > 0).all()
    assert (input_dataframe_by_entity['person'].loc[
        input_dataframe_by_entity['person']['household_role'] == "first_parent",
        'salary'
        ] <= salary_max_value).all()

    assert (input_dataframe_by_entity['household']['rent'] > 0).all()
    assert (input_dataframe_by_entity['household']['rent'] < rent_max_value).all()


# On vérifie que l'attribut `used_as_input_variables` correspond à la liste des variables
# qui sont employées dans le calcul des simulations, les autres variables n'étant pas utilisées dans le calcul,
# étant dans la base en entrée mais pas dans la base en sortie (la base de la simulation)
def test_init_from_data(nb_persons = 10, nb_groups = 5, salary_max_value = 50000,
        rent_max_value = 1000):

    # Set up test : the minimum necessary data to perform an `init_from_data`
    survey_scenario = AbstractSurveyScenario()
    assert survey_scenario.simulations is None
    # Generate some data and its period
    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value)
    period = periods.period('2017-01')
    # Creating a data object associated to its period, and we give it a name
    data_in = {
        'input_data_frame_by_entity_by_period': {
            period: input_data_frame_by_entity
            }
        }
    # data_in = copy.deepcopy(data_in) # Pour comparer avec la sortie de `init_from_data`
    table_ind = input_data_frame_by_entity['person'].copy(deep=True)
    table_men = input_data_frame_by_entity['household'].copy(deep=True)
    # print(table_ind)

    # We must add a TBS to the scenario to indicate what are the entities
    survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
    assert len(survey_scenario.tax_benefit_systems) == 1
    assert list(survey_scenario.tax_benefit_systems.keys()) == ["baseline"]
    assert survey_scenario.simulations is None
    # We must add the `used_as_input_variables` even though they don't seem necessary
    survey_scenario.used_as_input_variables = ['salary', 'rent', 'household_weight']
    # We must add the year to initiate a .new_simulation
    survey_scenario.period = 2017
    # Then we can input the data+period dict inside the scenario
    survey_scenario.init_from_data(data = data_in)
    assert len(survey_scenario.simulations) == 1
    # We are looking for the dataframes inside the survey_scenario
    all_var = list(set(list(table_ind.columns) + list(table_men.columns)))
    # print('Variables', all_var)
    data_out = survey_scenario.create_data_frame_by_entity(variables = all_var, period = period, merge = False)

    # 1 - Has the data object changed ? We only compare variables because Id's and others are lost in the process
    for cols in table_ind:
        if cols in data_out['person']:
            pass
        else:
            print('Columns lost in person table: ', cols)  # noqa T201
    assert data_out['person']['salary'].equals(table_ind['salary'])

    for cols in table_men:
        if cols in data_out['household']:
            pass
        else:
            print('Columns lost in household table: ', cols)  # noqa T201
    assert data_out['household']['rent'].equals(table_men['rent'])


def test_survey_scenario_input_dataframe_import(nb_persons = 10, nb_groups = 5, salary_max_value = 50000,
        rent_max_value = 1000):

    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value)
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
    survey_scenario.period = 2017
    survey_scenario.used_as_input_variables = ['salary', 'rent']
    period = periods.period('2017-01')
    data = {
        'input_data_frame_by_entity_by_period': {
            period: input_data_frame_by_entity
            },
        'config_files_directory': default_config_files_directory
        }
    survey_scenario.init_from_data(data = data)

    simulation = survey_scenario.simulations["baseline"]
    assert (
        simulation.calculate('salary', period) == input_data_frame_by_entity['person']['salary']
        ).all()
    assert (
        simulation.calculate('rent', period) == input_data_frame_by_entity['household']['rent']
        ).all()


def test_survey_scenario_input_dataframe_import_scrambled_ids(nb_persons = 10, nb_groups = 5, salary_max_value = 50000,
        rent_max_value = 1000):
    '''
        On teste que .init_from_data fait
    '''
    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value)  # Un dataframe d'exemple que l'on injecte
    input_data_frame_by_entity['person']['household_id'] = 4 - input_data_frame_by_entity['person']['household_id']
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
    survey_scenario.period = 2017
    survey_scenario.used_as_input_variables = ['salary', 'rent']
    period = periods.period('2017-01')
    data = {
        'input_data_frame_by_entity_by_period': {
            period: input_data_frame_by_entity
            },
        'config_files_directory': default_config_files_directory
        }
    survey_scenario.init_from_data(data = data)
    simulation = survey_scenario.simulations["baseline"]
    period = periods.period('2017-01')
    assert (
        simulation.calculate('salary', period) == input_data_frame_by_entity['person']['salary']
        ).all()
    assert (
        simulation.calculate('rent', period) == input_data_frame_by_entity['household']['rent']
        ).all()


def test_dump_survey_scenario():
    survey_scenario = create_randomly_initialized_survey_scenario()
    directory = os.path.join(
        openfisca_survey_manager_location,
        'openfisca_survey_manager',
        'tests',
        'data_files',
        'dump',
        )
    if os.path.exists(directory):
        shutil.rmtree(directory)

    survey_scenario.dump_simulations(directory = directory)
    period = "2017-01"
    df = survey_scenario.create_data_frame_by_entity(variables = ['salary', 'rent'], period = period)
    household = df['household']
    person = df['person']
    assert not household.empty
    assert not person.empty
    del survey_scenario
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
    survey_scenario.used_as_input_variables = ['salary', 'rent']
    survey_scenario.period = 2017
    survey_scenario.restore_simulations(directory = directory)
    df2 = survey_scenario.create_data_frame_by_entity(variables = ['salary', 'rent'], period = '2017-01')

    assert (df2['household'] == household).all().all()
    assert (df2['person'] == person).all().all()


@pytest.mark.order(before="test_add_survey_to_collection.py::test_add_survey_to_collection")
def test_inflate():
    survey_scenario = create_randomly_initialized_survey_scenario(collection = None)
    period = "2017-01"
    inflator = 2.42
    inflator_by_variable = {'rent': inflator}

    rent_before_inflate = survey_scenario.compute_aggregate('rent', period = period)
    survey_scenario.inflate(inflator_by_variable = inflator_by_variable, period = period)
    rent_after_inflate = survey_scenario.compute_aggregate('rent', period = period)
    assert_near(
        rent_after_inflate,
        inflator * rent_before_inflate,
        relative_error_margin = 1e-6,
        message = "Failing inflate with inflator_by_variable: rent_after_inflate = {} != {} = rent_before_inflate ({}) x inflator ({})".format(
            rent_after_inflate,
            rent_before_inflate * inflator,
            rent_before_inflate,
            inflator
            )
        )

    target = 3e5
    target_by_variable = {'salary': target}
    salary_before_inflate = survey_scenario.compute_aggregate('salary', period = period)
    survey_scenario.inflate(target_by_variable = target_by_variable, period = period)

    salary_after_inflate = survey_scenario.compute_aggregate('salary', period = period)
    assert_near(
        salary_after_inflate,
        target,
        relative_error_margin = 1e-6,
        message = "Failing inflate with inflator_by_variable: salary_after_inflate = {} != {} = target (salary_before_inflate = {})\n".format(
            salary_after_inflate,
            target,
            salary_before_inflate,
            )
        )


@pytest.mark.order(before="test_add_survey_to_collection.py::test_add_survey_to_collection")
def test_compute_pivot_table():
    survey_scenario = create_randomly_initialized_survey_scenario(collection = None)
    period = "2017-01"
    pivot_table = survey_scenario.compute_pivot_table(columns = ['age'], values = ["salary"], period = period, simulation = "baseline")

    assert pivot_table.index == "salary"
    assert pivot_table.values.round() == 21748

    del survey_scenario.weight_variable_by_entity
    survey_scenario.set_weight_variable_by_entity()
    pivot_table = survey_scenario.compute_pivot_table(columns = ['age'], values = ["salary"], period = period, simulation = "baseline")

    assert pivot_table.values.round() == 13570.


def test_compute_quantile():
    survey_scenario = create_randomly_initialized_survey_scenario()
    period = "2017-01"
    quintiles = survey_scenario.compute_quantiles(variable = "salary", nquantiles = 5, period = period, weighted = False, simulation = "baseline")
    return quintiles


if __name__ == "__main__":
    import sys
    log = logging.getLogger(__name__)
    logging.basicConfig(level = logging.DEBUG, stream = sys.stdout)
    quintiles = test_compute_quantile()
    # pivot_table = test_compute_pivot_table()
    # test_inflate()
    # test_create_data_frame_by_entity()



import logging
from pandas.core.frame import DataFrame

log = logging.getLogger(__name__)


def read_sas(sas_file_path, clean = False) -> DataFrame:
    try:
        import pyreadstat
        data_frame, _ = pyreadstat.read_sas7bdat(sas_file_path)
    except ImportError as e1:
        log.info("pyreadstat not available trying SAS7BDAT")
        try:
            from sas7bdat import SAS7BDAT
            data_frame = SAS7BDAT(sas_file_path).to_data_frame()
        except ImportError as e2:
            log.info("Neither pyreadstat nor SAS7BDAT are available")
            print(e1)  # noqa analysis:ignore
            raise e2

    return data_frame



import configparser
import os


from openfisca_survey_manager import default_config_files_directory as config_files_directory


def create_raw_data_ini(value_by_option_by_section = None):
    """Creates raw_data.ini configureation file

    Args:
      value_by_option_by_section(dict(dict)): Options value by section (Default value = None)

    """
    config_parser = configparser.ConfigParser()

    if value_by_option_by_section is not None:
        for section, value_by_option in value_by_option_by_section.items():
            config_parser.add_section(section)
            for option, value in value_by_option.items():
                config_parser.set(section, option, value)

    with open(os.path.join(config_files_directory, 'raw_data.ini'), 'w') as raw_data_config_file:
        config_parser.write(raw_data_config_file)





from numpy import argsort, asarray, cumsum, linspace, logical_and as and_, ones, repeat, zeros
import pandas as pd
import wquantiles
import weightedcalcs as wc
import numpy as np


def gini(values, weights = None):
    """Computes Gini coefficient (normalized to 1).
    # Using fastgini formula :
    #             i=N      j=i
    #             SUM W_i*(SUM W_j*X_j - W_i*X_i/2)
    #             i=1      j=1
    # G = 1 - 2* ----------------------------------
    #                 i=N             i=N
    #                 SUM W_i*X_i  *  SUM W_i
    #                 i=1             i=1
    # where observations are sorted in ascending order of X.
    # From http://fmwww.bc.edu/RePec/bocode/f/fastgini.html

    Args:
      values: Vector of values
      weights: Weights vector (Default value = None)

    Returns:
        float: Gini
    """
    if weights is None:
        weights = ones(len(values))

    df = pd.DataFrame({'x': values, 'w': weights})
    df = df.sort_values(by='x')
    x = df['x']
    w = df['w']
    wx = w * x

    cdf = cumsum(wx) - 0.5 * wx
    numerator = (w * cdf).sum()
    denominator = ((wx).sum()) * (w.sum())
    gini = 1 - 2 * (numerator / denominator)

    return gini


def kakwani(values, ineq_axis, weights = None):
    """Computes the Kakwani index

    Args:
      values: Vector of values
      ineq_axis: Inequality axis
      weights: Weights vector (Default value = None)

    Returns:
        float: Kakwani index
    """
    from scipy.integrate import simps

    if weights is None:
        weights = ones(len(values))

    PLCx, PLCy = pseudo_lorenz(values, ineq_axis, weights)
    LCx, LCy = lorenz(ineq_axis, weights)

    del PLCx

    return simps((LCy - PLCy), LCx)


def lorenz(values, weights = None):
    """Computes Lorenz curve coordinates (x, y)

    Args:
      values: Vector of values
      weights: Weights vector (Default value = None)

    Returns:
        (np.array, np.array): Lorenz curve coordinates
    """
    if weights is None:
        weights = ones(len(values))

    df = pd.DataFrame({'v': values, 'w': weights})
    df = df.sort_values(by = 'v')
    x = cumsum(df['w'])
    x = x / float(x[-1:])
    y = cumsum(df['v'] * df['w'])
    y = y / float(y[-1:])

    return x, y


def mark_weighted_percentiles(a, labels, weights, method, return_quantiles=False):
    """

    Args:
      a:
      labels:
      weights:
      method:
      return_quantiles:  (Default value = False)

    Returns:

    """
    # from http://pastebin.com/KTLip9ee
    # a is an input array of values.
    # weights is an input array of weights, so weights[i] goes with a[i]
    # labels are the names you want to give to the xtiles
    # method refers to which weighted algorithm.
    #      1 for wikipedia, 2 for the stackexchange post.

    # The code outputs an array the same shape as 'a', but with
    # labels[i] inserted into spot j if a[j] falls in x-tile i.
    # The number of xtiles requested is inferred from the length of 'labels'.

    np.random.seed(42)

    # First method, "vanilla" weights from Wikipedia article.
    if method == 1:

        # Sort the values and apply the same sort to the weights.
        N = len(a)
        sort_indx = argsort(a)
        tmp_a = a[sort_indx].copy()
        tmp_weights = weights[sort_indx].copy()

        # 'labels' stores the name of the x-tiles the user wants,
        # and it is assumed to be linearly spaced between 0 and 1
        # so 5 labels implies quintiles, for example.
        num_categories = len(labels)
        breaks = linspace(0, 1, num_categories + 1)

        # Compute the percentile values at each explicit data point in a.
        cu_weights = cumsum(tmp_weights)
        p_vals = (1.0 / cu_weights[-1]) * (cu_weights - 0.5 * tmp_weights)

        # Set up the output array.
        ret = repeat(0, len(a))
        if len(a) < num_categories:
            return ret

        # Set up the array for the values at the breakpoints.
        quantiles = []

        # Find the two indices that bracket the breakpoint percentiles.
        # then do interpolation on the two a_vals for those indices, using
        # interp-weights that involve the cumulative sum of weights.
        for brk in breaks:
            if brk <= p_vals[0]:
                i_low = 0
                i_high = 0
            elif brk >= p_vals[-1]:
                i_low = N - 1
                i_high = N - 1
            else:
                for ii in range(N - 1):
                    if (p_vals[ii] <= brk) and (brk < p_vals[ii + 1]):
                        i_low = ii
                        i_high = ii + 1

            if i_low == i_high:
                v = tmp_a[i_low]
            else:
                # If there are two brackets, then apply the formula as per Wikipedia.
                v = (
                    tmp_a[i_low]
                    + (
                        (brk - p_vals[i_low]) / (p_vals[i_high] - p_vals[i_low])) * (tmp_a[i_high] - tmp_a[i_low])
                    )

            # Append the result.
            quantiles.append(v)

        # Now that the weighted breakpoints are set, just categorize
        # the elements of a with logical indexing.
        for i in range(0, len(quantiles) - 1):
            lower = quantiles[i]
            upper = quantiles[i + 1]
            ret[and_(a >= lower, a < upper)] = labels[i]

        # make sure upper and lower indices are marked
        ret[a <= quantiles[0]] = labels[0]
        ret[a >= quantiles[-1]] = labels[-1]

        return ret

    # The stats.stackexchange suggestion.
    elif method == 2:

        N = len(a)
        sort_indx = argsort(a)
        tmp_a = a[sort_indx].copy()
        tmp_weights = weights[sort_indx].copy()

        num_categories = len(labels)
        breaks = linspace(0, 1, num_categories + 1)

        cu_weights = cumsum(tmp_weights)

        # Formula from stats.stackexchange.com post.
        s_vals = [0.0]
        for ii in range(1, N):
            s_vals.append(ii * tmp_weights[ii] + (N - 1) * cu_weights[ii - 1])
        s_vals = asarray(s_vals)

        # Normalized s_vals for comapring with the breakpoint.
        norm_s_vals = (1.0 / s_vals[-1]) * s_vals

        # Set up the output variable.
        ret = repeat(0, N)
        if N < num_categories:
            return ret

        # Set up space for the values at the breakpoints.
        quantiles = []

        # Find the two indices that bracket the breakpoint percentiles.
        # then do interpolation on the two a_vals for those indices, using
        # interp-weights that involve the cumulative sum of weights.
        for brk in breaks:
            if brk <= norm_s_vals[0]:
                i_low = 0
                i_high = 0
            elif brk >= norm_s_vals[-1]:
                i_low = N - 1
                i_high = N - 1
            else:
                for ii in range(N - 1):
                    if (norm_s_vals[ii] <= brk) and (brk < norm_s_vals[ii + 1]):
                        i_low = ii
                        i_high = ii + 1

            if i_low == i_high:
                v = tmp_a[i_low]
            else:
                # Interpolate as in the method 1 method, but using the s_vals instead.
                v = (
                    tmp_a[i_low]
                    + (
                        ((brk * s_vals[-1]) - s_vals[i_low])
                        / (s_vals[i_high] - s_vals[i_low])
                        ) * (tmp_a[i_high] - tmp_a[i_low])
                    )
            quantiles.append(v)

        # Now that the weighted breakpoints are set, just categorize
        # the elements of a as usual.
        for i in range(0, len(quantiles) - 1):
            lower = quantiles[i]
            upper = quantiles[i + 1]
            ret[and_(a >= lower, a < upper)] = labels[i]

        # make sure upper and lower indices are marked
        ret[a <= quantiles[0]] = labels[0]
        ret[a >= quantiles[-1]] = labels[-1]

        if return_quantiles:
            return ret, quantiles
        else:
            return ret


def pseudo_lorenz(values, ineq_axis, weights = None):
    """Computes The pseudo Lorenz Curve coordinates

    Args:
      values:
      ineq_axis:
      weights:  (Default value = None)

    Returns:

    """
    if weights is None:
        weights = ones(len(values))
    df = pd.DataFrame({'v': values, 'a': ineq_axis, 'w': weights})
    df = df.sort_values(by = 'a')
    x = cumsum(df['w'])
    x = x / float(x[-1:])
    y = cumsum(df['v'] * df['w'])
    y = y / float(y[-1:])

    return x, y


def bottom_share(values, rank_from_bottom, weights = None):
    """

    Args:
      values(np.array): Vector of values
      rank_from_bottom(float): Rank from bottom (bottom is 0 and top is 1)
      weights(np.array): Weights vector (Default value = None)

    Returns:

    """
    if weights is None:
        weights = ones(len(values))

    calc = wc.Calculator("weights")
    data_frame = pd.DataFrame({
        'weights': weights,
        'data': values,
        })
    quantile = calc.quantile(data_frame, 'data', rank_from_bottom)

    return (
        (data_frame["data"] < quantile)
        * data_frame["data"]
        * data_frame["weights"]
        ).sum() / (
            data_frame["data"]
            * data_frame["weights"]
            ).sum()


def top_share(values, rank_from_top, weights = None):
    """

    Args:
      values(np.array): Vector of values
      rank_from_top(float): Rank from top (bottom is 1 and top is 0)
      weights(np.array): Weights vector (Default value = None)

    Returns:

    """
    if weights is None:
        weights = ones(len(values))

    calc = wc.Calculator("weights")
    data_frame = pd.DataFrame({
        'weights': weights,
        'data': values,
        })
    quantile = calc.quantile(data_frame, 'data', 1 - rank_from_top)
    return (
        (data_frame["data"] >= quantile)
        * data_frame["data"]
        * data_frame["weights"]
        ).sum() / (
            data_frame["data"]
            * data_frame["weights"]
            ).sum()


def weighted_quantiles(data, labels, weights, return_quantiles = False):
    num_categories = len(labels)
    breaks = linspace(0, 1, num_categories + 1)
    quantiles = [
        wquantiles.quantile_1D(data, weights, mybreak) for mybreak in breaks[1:]
        ]
    ret = zeros(len(data))
    for i in range(0, len(quantiles) - 1):
        lower = quantiles[i]
        upper = quantiles[i + 1]
        ret[and_(data >= lower, data < upper)] = labels[i]

    if return_quantiles:
        return ret + 1, quantiles
    else:
        return ret + 1


def weightedcalcs_quantiles(data, labels, weights, return_quantiles = False):
    calc = wc.Calculator("weights")
    num_categories = len(labels)
    breaks = linspace(0, 1, num_categories + 1)
    data_frame = pd.DataFrame({
        'weights': weights,
        'data': data,
        })
    quantiles = [
        calc.quantile(data_frame, 'data', mybreak) for mybreak in breaks[1:]
        ]

    ret = zeros(len(data))
    for i in range(0, len(quantiles) - 1):
        lower = quantiles[i]
        upper = quantiles[i + 1]
        ret[and_(data > lower, data <= upper)] = labels[i]

    if return_quantiles:
        return ret + 1, quantiles
    else:
        return ret + 1



import logging
from numpy import arange

from openfisca_core.model_api import Variable, ADD, where, YEAR
from openfisca_survey_manager.statshelpers import (
    mark_weighted_percentiles,
    weightedcalcs_quantiles
    )

log = logging.getLogger(__name__)


def create_quantile(x, nquantiles, weight_variable, entity_name):
    class quantile(Variable):
        value_type = int
        entity = entity_name
        label = "Quantile"
        definition_period = YEAR

        def formula(entity, period):
            try:
                variable = entity(x, period)
            except ValueError as e:
                log.debug(f"Caught {e}")
                log.debug(f"Computing on whole period {period} via the ADD option")
                variable = entity(x, period, options = [ADD])

            weight = entity(weight_variable, period)
            labels = arange(1, nquantiles + 1)
            method = 2
            if len(weight) == 1:
                return weight * 0
            quantile, values = mark_weighted_percentiles(variable, labels, weight, method, return_quantiles = True)
            del values
            return quantile

    return quantile


def quantile(q, variable, weight_variable = None, filter_variable = None):
    """
    Return quantile of a variable with weight provided by a specific wieght variable potentially filtered
    """
    def formula(entity, period):
        value = entity(variable, period)
        if weight_variable is not None:
            weight = entity(weight_variable, period)
        weight = entity.filled_array(1)
        if filter_variable is not None:
            filter_value = entity(filter_variable, period)
            weight = filter_value * weight

        labels = arange(1, q + 1)
        quantile, _ = weightedcalcs_quantiles(
            value,
            labels,
            weight,
            return_quantiles = True,
            )
        if filter_variable is not None:
            quantile = where(weight > 0, quantile, -1)
        return quantile

    return formula


def old_quantile(q, variable, weight_variable = None, filter_variable = None):
    def formula(entity, period):
        value = entity(variable, period)
        if weight_variable is not None:
            weight = entity(weight_variable, period)
        weight = entity.filled_array(1)
        if filter_variable is not None:
            filter_value = entity(filter_variable, period)
            weight = filter_value * weight

        labels = arange(1, q + 1)
        quantile, _ = mark_weighted_percentiles(
            value,
            labels,
            weight,
            method = 2,  # * filter,
            return_quantiles = True,
            )
        if filter_variable is not None:
            quantile = where(weight > 0, quantile, -1)
        return quantile

    return formula



import os
import configparser


class Config(configparser.ConfigParser):
    config_ini = None

    def __init__(self, config_files_directory = None):
        configparser.ConfigParser.__init__(self)
        if config_files_directory is not None:
            config_ini = os.path.join(config_files_directory, 'config.ini')
            assert os.path.exists(config_ini), f"{config_ini} is not a valid path"
            self.config_ini = config_ini
            self.read([config_ini])

    def save(self):
        assert self.config_ini, "configuration file path is not defined"
        assert os.path.exists(self.config_ini)
        config_file = open(self.config_ini, 'w')
        self.write(config_file)
        config_file.close()



import logging

import numpy
from numpy import logical_not
import pandas as pd

from openfisca_core.model_api import Enum
from openfisca_survey_manager.calmar import calmar


log = logging.getLogger(__name__)


class Calibration(object):
    """An object to calibrate survey data of a SurveyScenario."""

    entity = None
    filter_by = None
    initial_entity_count = None
    _initial_weight_name = None
    initial_weight_by_entity = dict()
    target_margins = dict()
    margins_by_variable = dict()
    parameters = {
        'use_proportions': True,
        'initial_weight': None,
        'method': None,  # 'linear', 'raking ratio', 'logit'
        'up': None,
        'invlo': None
        }
    period = None
    simulation = None
    target_entity_count = None
    weight_name = None

    def __init__(self, simulation, target_margins, period, target_entity_count = None, parameters = None,
            filter_by = None, entity = None):
        self.period = period
        self.simulation = simulation
        if target_margins:
            margin_variables = list(target_margins.keys())
        else:
            margin_variables = []

        variable_instance_by_variable_name = simulation.tax_benefit_system.variables
        entities = set(
            variable_instance_by_variable_name[variable].entity.key
            for variable in margin_variables
            )

        if len(entities) == 0:
            assert target_entity_count != 0
            assert entity in [
                entity.key
                for entity in simulation.tax_benefit_system.entities
                ]
        elif len(entities) > 1:
            raise NotImplementedError("Cannot hande multiple entites")
        else:
            entity = list(entities)[0]

        assert simulation.weight_variable_by_entity is not None

        weight_variable_name = simulation.weight_variable_by_entity.get(entity)
        self.weight_name = weight_name = weight_variable_name

        self.entity = entity
        period = self.period

        if filter_by:
            self.filter_by = simulation.calculate(filter_by, period = period)
        else:
            self.filter_by = numpy.array(1.0)

        assert weight_name is not None, "A calibration needs a weight variable name to act on"
        self._initial_weight_name = weight_name + "_ini"
        self.initial_weight = initial_weight = simulation.calculate(weight_name, period = period)

        self.initial_entity_count = sum(initial_weight * self.filter_by)
        self.target_entity_count = target_entity_count

        self.weight = initial_weight.copy()

        # TODO does not work
        for entity, weight_variable in simulation.weight_variable_by_entity.items():
            self.initial_weight_by_entity[entity] = simulation.calculate(weight_variable, period = period)

        if target_margins:
            for variable, target in target_margins.items():
                self.set_target_margin(variable, target)

        self.parameters = parameters

    def _build_calmar_data(self) -> pd.DataFrame:
        """Build the data dictionnary used as calmar input argument.

        Returns:
            pd.DataFrame: Data used by calmar
        """
        # Select only filtered entities
        assert self._initial_weight_name is not None
        data = pd.DataFrame()
        data[self._initial_weight_name] = self.initial_weight * self.filter_by
        period = self.period
        for variable in self.margins_by_variable:
            assert variable in self.simulation.tax_benefit_system.variables
            data[variable] = self.simulation.calculate(variable, period = period)

        return data

    def calibrate(self, inplace = False):
        """Apply the calibrations by updating weights and margins.

        Args:
            inplace (bool, optional): Whether to return the calibrated or to setthem inplace. Defaults to False.

        Returns:
            numpy.array: calibrated weights
        """
        assert self.margins_by_variable is not None, "Margins by variable should be set"
        margins_by_variable = self.margins_by_variable
        parameters = self.get_parameters()

        if margins_by_variable is not None:
            simple_margins_by_variable = dict([
                (variable, margins_by_type['target'])
                for variable, margins_by_type in margins_by_variable.items()])
        else:
            simple_margins_by_variable = dict()

        if self.target_entity_count:
            simple_margins_by_variable['total_population'] = self.target_entity_count

        self._update_weights(simple_margins_by_variable, parameters = parameters)
        self._update_margins()
        if inplace:
            self.set_calibrated_weights()
            return

        return self.weight

    def get_parameters(self) -> dict:
        """Get the parameters.

        Returns:
            dict: Parameters
        """
        p = {}
        p['method'] = self.parameters.get('method', 'linear')
        if self.parameters.get('invlo') is not None:
            p['lo'] = 1 / self.parameters.get('invlo')
        p['up'] = self.parameters.get('up')
        if p['method'] == 'logit':
            assert self.parameters.get('invlo') is not None and self.parameters.get('up') is not None
            p['lo'] = 1 / self.parameters.get('invlo')
            p['up'] = self.parameters.get('up')
        p['use_proportions'] = self.parameters.get('use_proportions', True)
        p['initial_weight'] = self.weight_name + ""
        return p

    def set_target_margin(self, variable, target):
        """Set variable target margin.

        Args:
          variable: Target variable
          target: Target value
        """
        simulation = self.simulation
        period = self.period
        assert variable in simulation.tax_benefit_system.variables
        variable_instance = simulation.tax_benefit_system.variables[variable]

        filter_by = self.filter_by
        target_by_category = None
        categorical_variable = (
            (variable_instance.value_type in [bool, Enum])
            or (variable_instance.unit == 'years')
            or (variable_instance.unit == 'months')
            )
        if categorical_variable:
            value = simulation.calculate(variable, period = period)
            filtered_value = value if all(filter_by) else value[filter_by.astype(bool)]
            categories = numpy.sort(numpy.unique(filtered_value))
            target_by_category = dict(zip(categories, target))

        if not self.margins_by_variable:
            self.margins_by_variable = dict()
        if variable not in self.margins_by_variable:
            self.margins_by_variable[variable] = dict()
        self.margins_by_variable[variable]['target'] = target_by_category or target
        self._update_margins()

    def reset(self):
        """Reset the calibration to its initial state."""
        simulation = self.simulation
        simulation.delete_arrays(self.weight_name, self.period)
        simulation.set_input(self.weight_name, self.period, numpy.array(self.initial_weight))

    def set_calibrated_weights(self):
        """Modify the weights to use the calibrated weights."""
        period = self.period
        simulation = self.simulation
        simulation.set_input(self.weight_name, period, self.weight)
        for weight_name in simulation.weight_variable_by_entity.values():
            if weight_name == self.weight_name:
                continue
            # Delete other entites already computed weigths
            # to ensure that this weights a recomputed if they derive from
            # the calibrated weight variable
            weight_variable = simulation.tax_benefit_system.variables[weight_name]
            if weight_variable.formulas:
                simulation.delete_arrays(weight_variable.name, period)

    def summary(self):
        """Summarize margins."""
        margins_df = pd.DataFrame.from_dict(self.margins_by_variable).T
        margins_df.loc['entity_count', 'actual'] = (self.weight * self.filter_by).sum()
        margins_df.loc['entity_count', 'initial'] = (self.initial_weight * self.filter_by).sum()
        margins_df.loc['entity_count', 'target'] = self.target_entity_count
        return margins_df

    def _update_margins(self):
        """Update margins."""
        for variable in self.margins_by_variable:
            simulation = self.simulation
            period = self.period
            entity = self.entity

            # These are the varying weights
            weight = self.weight
            filter_by = self.filter_by
            initial_weight = self.initial_weight

            value = simulation.calculate(variable, period = period)
            weight_variable = simulation.weight_variable_by_entity[entity]

            # weight = simulation.calculate(weight_variable, period)
            # initial_weight = self.initial_weight_by_entity[entity]

            if filter_by != 1:
                if weight_variable != self.weight_name:
                    NotImplementedError("No filtering possible so far when target varaible is not on the same entity as varying weights")

                weight = weight[filter_by]
                initial_weight = initial_weight[filter_by]
                value = value[filter_by]

            margin_items = [
                ('actual', weight),
                ('initial', initial_weight),
                ]

            variable_instance = simulation.tax_benefit_system.get_variable(variable)
            assert variable_instance is not None
            if variable_instance.value_type in [bool, Enum]:
                margin_items.append(('category', value))
                margins_data_frame = pd.DataFrame.from_items(margin_items)
                margins_data_frame = margins_data_frame.groupby('category', sort = True).sum()
                margin_by_type = margins_data_frame.to_dict()
            else:
                margin_by_type = dict(
                    actual = (weight * value).sum(),
                    initial = (initial_weight * value).sum(),
                    )
            self.margins_by_variable[variable].update(margin_by_type)

    def _update_weights(self, margins, parameters = None):
        """Run calmar, stores new weights and returns adjusted margins.

        Args:
          margins: margins
          parameters:  Parameters (Default value = {})

        Returns:
            dict: Updated margins

        """
        if parameters is None:
            parameters = dict()

        entity = self.entity
        weight_variable = self.simulation.weight_variable_by_entity[entity]

        if self.weight_name != weight_variable:
            raise NotImplementedError("Calmar needs to be adapted. Consider using a projected target on the entity with changing weights")

        data = self._build_calmar_data()

        assert self._initial_weight_name is not None
        parameters['initial_weight'] = self._initial_weight_name
        if self.target_entity_count:
            margins["total_population"] = self.target_entity_count

        val_pondfin, lambdasol, updated_margins = calmar(
            data, margins, **parameters)
        # Updating only afetr filtering weights
        self.weight = val_pondfin * self.filter_by + self.weight * (logical_not(self.filter_by))

        return updated_margins



import pytest

from openfisca_core.tools import assert_near
from openfisca_core import periods

from openfisca_survey_manager.tests.test_scenario import create_randomly_initialized_survey_scenario
from openfisca_survey_manager.tests.test_scenario import generate_input_input_dataframe_by_entity
from openfisca_survey_manager.calibration import Calibration

from openfisca_survey_manager import default_config_files_directory
from openfisca_survey_manager.scenarios.abstract_scenario import AbstractSurveyScenario
from openfisca_survey_manager.tests import tax_benefit_system


def test_calibration_variable_entity_is_weight_entity():
    survey_scenario = create_randomly_initialized_survey_scenario(collection=None)
    period = "2017-01"
    survey_scenario.period = period
    person_weight_before = survey_scenario.calculate_series("person_weight", period)
    # 'initial_rent_aggregate' is assigned to but never used initial_rent_aggregate = survey_scenario.compute_aggregate("rent", period = period)
    target_rent_aggregate = 200000

    survey_scenario.calibrate(
        period,
        target_margins_by_variable = {
            'rent': target_rent_aggregate,
            },
        parameters = {"method": "raking ratio"},
        )

    for _, simulation in survey_scenario.simulations.items():
        assert all(simulation.calibration.weight != simulation.calibration.initial_weight)

    assert_near(survey_scenario.compute_aggregate("rent", period = period), target_rent_aggregate)

    # See if propagation to derived weights is done well
    person_weight_after = survey_scenario.calculate_series("person_weight", period)
    assert (person_weight_after != person_weight_before).all()


def test_calibration_variable_entity_is_not_weight_entity():
    survey_scenario = create_randomly_initialized_survey_scenario(collection=None)
    period = "2017-01"
    survey_scenario.period = period
    target_rent_aggregate = 200000
    target_salary_aggregate = 1e7

    with pytest.raises(NotImplementedError):
        survey_scenario.calibrate(
            period,
            target_margins_by_variable = {
                'rent': target_rent_aggregate,
                'salary': target_salary_aggregate
                },
            parameters = {"method": "raking ratio"},
            )


def test_simulation_calibration_variable_entity_is_weight_entity():
    survey_scenario = create_randomly_initialized_survey_scenario(collection=None)
    period = "2017-01"
    survey_scenario.period = period
    simulation = list(survey_scenario.simulations.values())[0]
    person_weight_before = simulation.calculate("person_weight", period)

    # initial_rent_aggregate = simulation.compute_aggregate("rent", period = period)
    target_rent_aggregate = 200000

    calibration = Calibration(
        simulation,
        period = "2017-01",
        target_margins = {
            'rent': target_rent_aggregate,
            },
        target_entity_count = 300,
        parameters = {"method": "raking ratio"},
        )

    calibration.calibrate(inplace = True)
    assert all(calibration.weight != calibration.initial_weight)

    assert_near(simulation.compute_aggregate("rent", period = period), target_rent_aggregate)

    # See if propagation to derived weights is done well
    person_weight_after = simulation.calculate("person_weight", period)
    assert all(person_weight_after != person_weight_before)
    assert calibration.initial_entity_count != calibration.target_entity_count
    assert simulation.calculate("household_weight", period).sum() == calibration.target_entity_count


def test_simulation_calibration_input_from_data():
    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        10, 5, 5000, 1000)
    survey_scenario = AbstractSurveyScenario()
    weight_variable_by_entity = {
        "person": "person_weight",
        "household": "household_weight",
        }
    survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
    survey_scenario.period = '2017-01'
    survey_scenario.used_as_input_variables = ['salary', 'rent', 'household_weight']
    period = periods.period('2017-01')
    target_rent_aggregate = 200000

    data = {
        'input_data_frame_by_entity_by_period': {
            period: input_data_frame_by_entity
            },
        'config_files_directory': default_config_files_directory
        }
    calibration_kwargs = {'target_margins_by_variable': {'rent': target_rent_aggregate}, 'target_entity_count': 300, 'parameters': {'method': 'logit', 'up': 4, 'invlo': 4}}
    survey_scenario.set_weight_variable_by_entity(weight_variable_by_entity)
    assert survey_scenario.weight_variable_by_entity == weight_variable_by_entity
    survey_scenario.init_from_data(data = data, calibration_kwargs=calibration_kwargs)
    for simulation_name, simulation in survey_scenario.simulations.items():
        assert simulation.weight_variable_by_entity == weight_variable_by_entity, f"{simulation_name} weight_variable_by_entity does not match {weight_variable_by_entity}"
        assert (survey_scenario.calculate_series("household_weight", period, simulation = simulation_name) != 0).all()
    return survey_scenario



import logging
import os
from pathlib import Path

log = logging.getLogger(__name__)


openfisca_survey_manager_location = Path(__file__).parent.parent


# Hack for use at the CASD (shared user)
# Use taxipp/.config/ directory if exists as default_config_files_directory
try:
    import taxipp
    taxipp_location = Path(taxipp.__file__).parent.parent
    default_config_files_directory = os.path.join(taxipp_location, '.config', 'openfisca-survey-manager')
except ImportError:
    taxipp_location = None

if taxipp_location is None or not os.path.exists(default_config_files_directory):
    default_config_files_directory = None


# Hack for using with france-data on a CI or locally
try:
    import openfisca_france_data
    france_data_location = Path(openfisca_france_data.__file__).parent.parent
    from xdg import BaseDirectory
    default_config_files_directory = BaseDirectory.save_config_path('openfisca-survey-manager')
except ImportError:
    france_data_location = None

if france_data_location is None or not os.path.exists(default_config_files_directory):
    default_config_files_directory = None

# Run CI when testing openfisca-survey-manager for example GitHub Actions
test_config_files_directory = os.path.join(
    openfisca_survey_manager_location,
    'openfisca_survey_manager',
    'tests',
    'data_files',
    )

with open(os.path.join(test_config_files_directory, 'config_template.ini')) as file:
    config_ini = file.read()

config_ini = config_ini.format(location = openfisca_survey_manager_location)
try:
    with open(os.path.join(test_config_files_directory, 'config.ini'), "w+") as file:
        file.write(config_ini)
except PermissionError:
    log.debug(f"config.ini can't be written in the test config files directory{test_config_files_directory}")
    pass

# GitHub Actions test
is_in_ci = 'CI' in os.environ
private_run_with_data = False

if is_in_ci and default_config_files_directory is None:
    if "CI_RUNNER_TAGS" in os.environ:
        private_run_with_data = (
            ("data-in" in os.environ["CI_RUNNER_TAGS"])
            # or ("data-out" in os.environ["CI_RUNNER_TAGS"])
            )
    if not private_run_with_data:
        default_config_files_directory = test_config_files_directory

if default_config_files_directory is None:
    from xdg import BaseDirectory
    default_config_files_directory = BaseDirectory.save_config_path('openfisca-survey-manager')

    log.debug(f'Using default_config_files_directory = {default_config_files_directory}')



#! /usr/bin/env python


import collections
import os
import re
import glob
import logging
import pandas
import yaml
import pyarrow.parquet as pq
import pyarrow as pa

from .tables import Table


ident_re = re.compile(r"(?i)ident\d{2,4}$")  # noqa

log = logging.getLogger(__name__)


source_format_by_extension = dict(
    csv = 'csv',
    sas7bdat = "sas",
    dta = 'stata',
    Rdata = 'Rdata',
    spss = 'sav',
    parquet = 'parquet',
    )

admissible_source_formats = list(source_format_by_extension.values())


class NoMoreDataError(Exception):
    # Exception when the user ask for more data than available in file
    pass


class Survey(object):
    """An object to describe survey data"""
    hdf5_file_path = None
    parquet_file_path = None
    informations = dict()
    label = None
    name = None
    tables = collections.OrderedDict()
    tables_index = dict()
    survey_collection = None

    def __init__(self, name = None, label = None, hdf5_file_path = None, parquet_file_path = None,
            survey_collection = None, **kwargs):
        assert name is not None, "A survey should have a name"
        self.name = name
        self.tables = dict()

        if label is not None:
            self.label = label

        if hdf5_file_path is not None:
            self.hdf5_file_path = hdf5_file_path

        if parquet_file_path is not None:
            self.parquet_file_path = parquet_file_path

        if survey_collection is not None:
            self.survey_collection = survey_collection

        self.informations = kwargs

    def __repr__(self):
        header = f"""{self.name} : survey data {self.label}
Contains the following tables : \n"""
        tables = yaml.safe_dump(
            list(self.tables.keys()),
            default_flow_style = False)
        informations = yaml.safe_dump(self.informations, default_flow_style = False)
        return header + tables + informations

    @classmethod
    def create_from_json(cls, survey_json):
        self = cls(
            name = survey_json.get('name'),
            label = survey_json.get('label'),
            hdf5_file_path = survey_json.get('hdf5_file_path'),
            parquet_file_path = survey_json.get('parquet_file_path'),
            **survey_json.get('informations', dict())
            )
        self.tables = survey_json.get('tables')
        return self

    def dump(self):
        assert self.survey_collection is not None
        self.survey_collection.dump()

    def fill_store(self, source_format = None, tables = None, overwrite = True, keep_original_parquet_file = False,
            store_format = "hdf5"):
        """
        Convert data from the source files to store format either hdf5 or parquet.
        If the source is in parquet, the data is not converted.
        """
        assert self.survey_collection is not None
        assert isinstance(overwrite, bool) or isinstance(overwrite, list)
        survey = self
        # Create folder if it does not exist
        config = survey.survey_collection.config
        directory_path = config.get("data", "output_directory")
        if not os.path.isdir(directory_path):
            log.warn(f"{directory_path} who should be the store data directory does not exist: we create the directory")
            os.makedirs(directory_path)

        if source_format == "parquet":
            store_format = "parquet"

        if store_format == "hdf5" and survey.hdf5_file_path is None:
            survey.hdf5_file_path = os.path.join(directory_path, survey.name + '.h5')

        if store_format == "parquet" and survey.parquet_file_path is None:
            survey.parquet_file_path = os.path.join(directory_path, survey.name)

        self.store_format = store_format

        if source_format is not None:
            assert source_format in admissible_source_formats, f"Data source format {source_format} is unknown"
            source_formats = [source_format]
        else:
            source_formats = admissible_source_formats

        for source_format in source_formats:
            files = f"{source_format}_files"
            for data_file in survey.informations.get(files, []):
                path_name, extension = os.path.splitext(data_file)
                name = os.path.basename(path_name)
                if tables is None or name in tables:
                    if keep_original_parquet_file:
                        # Use folder instead of files if numeric at end of file
                        if re.match(r".*-\d$", name):
                            name = name.split("-")[0]
                            parquet_file = os.path.dirname(data_file)
                            # Get the parent folder
                            survey.parquet_file_path = "/".join(os.path.dirname(data_file).split(os.sep)[:-1])
                        else:
                            parquet_file = data_file
                            survey.parquet_file_path = os.path.dirname(data_file)
                        table = Table(
                            label = name,
                            name = name,
                            source_format = source_format_by_extension[extension[1:]],
                            survey = survey,
                            parquet_file = parquet_file,
                            )
                        table.read_parquet_columns(data_file)

                    else:
                        table = Table(
                            label = name,
                            name = name,
                            source_format = source_format_by_extension[extension[1:]],
                            survey = survey,
                            )
                        table.fill_store(
                            data_file,
                            clean = True,
                            overwrite = overwrite if isinstance(overwrite, bool) else table.name in overwrite,
                            )
        self.dump()

    def find_tables(self, variable, tables = None, rename_ident = True):
        """Find tables containing a given variable."""
        container_tables = []

        assert variable is not None

        if tables is None:
            tables = self.tables
        tables_index = self.tables_index
        for table in tables:
            if table not in tables_index:
                tables_index[table] = self.get_columns(table)
            if variable in tables_index[table]:
                container_tables.append(table)
        return container_tables

    def get_columns(self, table, rename_ident = True):
        """
        Get columns of a table.
        """
        assert table is not None
        if self.hdf5_file_path is not None:
            store = pandas.HDFStore(self.hdf5_file_path, "r")
            if table in store:
                log.debug(f"Building columns index for table {table}")
                data_frame = store[table]
                if rename_ident is True:
                    for column_name in data_frame:
                        if ident_re.match(column_name) is not None:
                            data_frame.rename(columns = {column_name: "ident"}, inplace = True)
                            log.info(f"{column_name} column have been replaced by ident")
                            break
                store.close()
                return list(data_frame.columns)
            else:
                log.info(f'table {table} was not found in {store.filename}')
                store.close()
                return list()
        elif self.parquet_file_path is not None:
            parquet_schema = pq.read_schema(self.parquet_file_path)
            column_names = parquet_schema.names
            return column_names

    def get_value(self, variable, table, lowercase = False, ignorecase = False):
        """Get variable value from a survey table.

        Args:
          variable: variable to retrieve
          table(str): name of the table
          lowercase(bool, optional, optional): lowercase variable names, defaults to False
          ignorecase: ignore case of table name, defaults to False

        Returns:
          pd.DataFrame: dataframe containing the variable

        """
        return self.get_values([variable], table)

    def get_values(self, variables = None, table = None, lowercase = False, ignorecase = False, rename_ident = True, batch_size = None, batch_index=0, filter_by=None) -> pandas.DataFrame:
        """Get variables values from a survey table.

        Args:
          variables(list, optional, optional): variables to retrieve, defaults to None (retrieve all variables)
          table(str, optional, optional): name of the table, defaults to None
          ignorecase: ignore case of table name, defaults to False
          lowercase(bool, optional, optional): lowercase variable names, defaults to False
          rename_ident(bool, optional, optional): rename ident+yr (e.g. ident08) into ident, defaults to True
          batch_size(int, optional, optional): batch size for parquet file, defaults to None
          batch_index(int, optional, optional): batch index for parquet file, defaults to 0

        Returns:
          pd.DataFrame: dataframe containing the variables

        Raises:
          Exception:

        """
        if self.parquet_file_path is None and self.hdf5_file_path is None:
            raise Exception(f"No data file found for survey {self.name}")
        if self.hdf5_file_path is not None:
            assert os.path.exists(self.hdf5_file_path), f'{self.hdf5_file_path} is not a valid path. This could happen because your data were not builded yet. Please consider using a rebuild option in your code.'
            store = pandas.HDFStore(self.hdf5_file_path, "r")
            if ignorecase:
                keys = store.keys()
                eligible_tables = []
                for string in keys:
                    match = re.findall(table, string, re.IGNORECASE)
                    if match:
                        eligible_tables.append(match[0])
                if len(eligible_tables) > 1:
                    raise ValueError(f"{table} is ambiguious since the following tables are available: {eligible_tables}")
                elif len(eligible_tables) == 0:
                    raise ValueError(f"No eligible available table in {keys}")
                else:
                    table = eligible_tables[0]
            try:
                df = store.select(table)
            except KeyError:
                log.error(f'No table {table} in the file {self.hdf5_file_path}')
                log.error(f'This could happen because your data were not builded yet. Available tables are: {store.keys()}')
                store.close()
                raise

            store.close()

        elif self.parquet_file_path is not None:
            if table is None:
                raise Exception("A table name is needed to retrieve data from a parquet file")
            for table_name, table_content in self.tables.items():
                if table_name in table:
                    parquet_file = table_content.get("parquet_file")
                    # Is parquet_file a folder or a file?
                    if os.path.isdir(parquet_file):
                        # find first parquet file in folder
                        for file in os.listdir(parquet_file):
                            if file.endswith('.parquet'):
                                one_parquet_file = os.path.join(parquet_file, file)
                                break
                        else:
                            raise Exception(f"No parquet file found in {parquet_file}")
                    else:
                        one_parquet_file = parquet_file
                    parquet_schema = pq.read_schema(one_parquet_file)
                    assert len(parquet_schema.names) >= 1, f"The parquet file {table_content.get('parquet_file')} is empty"
                    columns = table_content.get('variables')
                    if filter_by:
                        df = pq.ParquetDataset(parquet_file, filters=filter_by).read(columns=columns).to_pandas()
                    elif batch_size:
                        if os.path.isdir(parquet_file):
                            parquet_file = glob.glob(os.path.join(parquet_file, '*.parquet'))
                        else:
                            parquet_file = [parquet_file]
                        # Initialize an empty list to store the Parquet tables
                        tables = []
                        # Loop through the file paths and read each Parquet file
                        for file_path in parquet_file:
                            table = pq.read_table(file_path, columns=columns)
                            tables.append(table)

                        # Concatenate the tables if needed
                        if len(tables) > 1:
                            final_table = pa.concat_tables(tables)
                        else:
                            final_table = tables[0]
                        record_batches = final_table.to_batches(max_chunksize=batch_size)
                        if len(record_batches) <= batch_index:
                            raise NoMoreDataError(f"Batch {batch_index} not found in {table_name}. Max index is {len(record_batches)}")
                        df = record_batches[batch_index].to_pandas()
                        # iter_parquet = parquet_file.iter_batches(batch_size=batch_size, columns=variables)
                        # index = 0
                        # while True:
                        #     try:
                        #         batch = next(iter_parquet)
                        #     except StopIteration:
                        #         raise NoMoreDataError(f"Batch {batch_index} not found in {table_name}. Max index is {index}")
                        #         break
                        #     if batch_index == index:
                        #         df = batch.to_pandas()
                        #         break
                        #     index += 1
                    else:
                        df = pq.ParquetDataset(parquet_file).read(columns=columns).to_pandas()
                    break
            else:
                raise Exception(f"No table {table} found in {self.parquet_file_path}")

        if lowercase:
            columns = dict((column_name, column_name.lower()) for column_name in df)
            df.rename(columns = columns, inplace = True)

        if rename_ident is True:
            for column_name in df:
                if ident_re.match(str(column_name)) is not None:
                    df.rename(columns = {column_name: "ident"}, inplace = True)
                    log.info(f"{column_name} column have been replaced by ident")
                    break

        if variables is None:
            return df
        else:
            diff = set(variables) - set(df.columns)
            if diff:
                raise Exception(f"The following variable(s) {diff} are missing")
            variables = list(set(variables).intersection(df.columns))
            df = df[variables]
            return df

    def insert_table(self, label = None, name = None, **kwargs):
        """
        Inserts a table in the Survey object
        If a pandas dataframe is provided, it is saved in the store file
        """
        parquet_file = kwargs.pop('parquet_file', None)
        data_frame = kwargs.pop('data_frame', None)
        if data_frame is None:
            # Try without underscore
            data_frame = kwargs.pop('dataframe', None)

        if data_frame is not None:
            assert isinstance(data_frame, pandas.DataFrame)
            variables = kwargs.pop('variables', None)
            if variables is not None:
                assert set(variables) < set(data_frame.columns)
            else:
                variables = list(data_frame.columns)
            if label is None:
                label = name
            table = Table(label = label, name = name, survey = self, variables = variables, parquet_file = parquet_file)
            assert (table.survey.hdf5_file_path is not None) or (table.survey.parquet_file_path is not None)
            if parquet_file is not None:
                log.debug(f"Saving table {name} in {table.survey.parquet_file_path}")
                data_frame.to_parquet(parquet_file)
            else:
                log.debug(f"Saving table {name} in {table.survey.hdf5_file_path}")
                to_hdf_kwargs = kwargs.pop('to_hdf_kwargs', dict())
                table.save_data_frame_to_hdf5(data_frame, **to_hdf_kwargs)

        if name not in self.tables:
            self.tables[name] = dict()
        for key, val in kwargs.items():
            self.tables[name][key] = val

    def to_json(self):
        self_json = collections.OrderedDict((
            ))
        self_json['hdf5_file_path'] = self.hdf5_file_path
        self_json['parquet_file_path'] = self.parquet_file_path
        self_json['label'] = self.label
        self_json['name'] = self.name
        self_json['tables'] = self.tables
        self_json['informations'] = collections.OrderedDict(sorted(self.informations.items()))
        return self_json



"""Tables."""

from chardet.universaldetector import UniversalDetector
import collections
import csv
import datetime
import errno
import gc
import logging
import os
import pandas
from pyarrow import parquet as pq


from openfisca_survey_manager import read_sas

try:
    from openfisca_survey_manager.read_spss import read_spss
except ImportError:
    read_spss = None


log = logging.getLogger(__name__)


reader_by_source_format = dict(
    # Rdata = pandas.rpy.common.load_data,
    csv = pandas.read_csv,
    sas = read_sas.read_sas,
    spss = read_spss,
    stata = pandas.read_stata,
    parquet = pandas.read_parquet,
    )


class Table(object):
    """A table of a survey."""
    label = None
    name = None
    source_format = None
    survey = None
    variables = None
    parquet_file = None

    def __init__(self, survey = None, name = None, label = None, source_format = None, variables = None, parquet_file = None,
                 **kwargs):
        assert name is not None, "A table should have a name"
        self.name = name
        self.label = label
        self.source_format = source_format
        self.variables = variables
        self.parquet_file = parquet_file
        self.informations = kwargs

        from .surveys import Survey  # Keep it here to avoid infinite recursion
        assert isinstance(survey, Survey), f'survey is of type {type(survey)} and not {Survey}'
        self.survey = survey
        if not survey.tables:
            survey.tables = collections.OrderedDict()

        survey.tables[name] = collections.OrderedDict(
            source_format = source_format,
            variables = variables,
            parquet_file = parquet_file,
            )

    def _check_and_log(self, data_file_path, store_file_path):
        """
        Check if the file exists and log the insertion.

        Args:
            data_file_path: Data file path
            store_file_path: Store file or dir path

        Raises:
            Exception: File not found
        """
        assert store_file_path is not None, "Store file path cannot be None"
        if not os.path.isfile(data_file_path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), data_file_path)

        log.info(f"Inserting table {self.name} from file {data_file_path} in store file {store_file_path} at point {self.name}")

    def _is_stored(self):
        if self.survey.hdf5_file_path is not None:
            store = pandas.HDFStore(self.survey.hdf5_file_path)
            if self.name in store:
                log.info(f'Exiting without overwriting {self.name} in {self.survey.hdf5_file_path}')
                store.close()
                return True

            store.close()
            return False
        else:
            return False

    def _save(self, data_frame: pandas.DataFrame = None, store_format = "hdf5"):
        """
        Save a data frame in the store according to is format (HDF5 or Parque).
        """
        assert data_frame is not None
        variables = self.variables

        if variables:
            stored_variables = list(set(variables).intersection(set(data_frame.columns)))
            log.info(f'The folloging variables are stored: {stored_variables}')
            if set(stored_variables) != set(variables):
                log.info(f'variables wanted by the user that were not available: {list(set(variables) - set(stored_variables))}')
            data_frame = data_frame[stored_variables].copy()

        assert store_format in ["hdf5", "parquet"], f"invalid store_format: {store_format}"
        if store_format == "hdf5":
            self.save_data_frame_to_hdf5(data_frame)
        else:
            parquet_file_path = self.survey.parquet_file_path
            log.info(f"Inserting table {self.name} in Parquet file {parquet_file_path}")
            self.save_data_frame_to_parquet(data_frame)
        gc.collect()

    def fill_store(self, data_file, overwrite: bool = False, clean: bool = False, **kwargs):
        """
        Fill the store (HDF5 or parquet file) with the table.
        Read the `data_file` in parameter and save it to the store.

        Args:
            data_file (_type_, optional): The data file path. Defaults to None.
            overwrite (bool, optional): Overwrite the data. Defaults to False.
            clean (bool, optional): Clean the raw data befoe saving. Defaults to False.
            store_format (str, optional): _description_. Defaults to "hdf5".

        Raises:
            e: Skip file if error
        """
        if not overwrite and self._is_stored():
            log.info(
                f'Exiting without overwriting {self.name} in {self.survey.hdf5_file_path}'
                )
            return

        start_table_time = datetime.datetime.now()
        data_frame = self.read_source(data_file, **kwargs)
        try:
            if clean:
                clean_data_frame(data_frame)
            self._save(data_frame = data_frame, store_format = self.survey.store_format)
            log.info(f"File {data_file} has been processed in {datetime.datetime.now() - start_table_time}")
        except Exception as e:
            log.info(f'Skipping file {data_file} because of following error \n {e}')
            raise e

    def read_parquet_columns(self, parquet_file = None) -> list:
        """
        Initialize the table from a parquet file.
        """
        if parquet_file is None:
            parquet_file = self.parquet_file
        log.info(f"Initializing table {self.name} from parquet file {parquet_file}")
        self.source_format = 'parquet'
        parquet_schema = pq.read_schema(parquet_file)
        self.variables = parquet_schema.names
        self.survey.tables[self.name]["variables"] = self.variables
        return self.variables

    def read_source(self, data_file, **kwargs):
        source_format = self.source_format
        store_file_path = (
            self.survey.hdf5_file_path
            if self.survey.store_format == "hdf5"
            else self.survey.parquet_file_path
            )

        self._check_and_log(data_file, store_file_path = store_file_path)
        reader = reader_by_source_format[source_format]
        try:
            if source_format == 'csv':
                try:
                    data_frame = reader(data_file, **kwargs)

                    if len(data_frame.columns) == 1 and ";" in len(data_frame.columns[0]):
                        raise ValueError("A ';' is present in the unique column name. Looks like we got the wrong separator.")

                except Exception:
                    log.debug(f"Failing to read {data_file}, Trying to infer encoding and dialect/separator")

                    # Detect encoding
                    detector = UniversalDetector()
                    with open(data_file, 'rb') as csvfile:
                        for line in csvfile:
                            detector.feed(line)
                            if detector.done:
                                break
                        detector.close()

                    encoding = detector.result['encoding']
                    confidence = detector.result['confidence']

                    # Sniff dialect
                    try:
                        with open(data_file, 'r', newline = "", encoding = encoding) as csvfile:
                            dialect = csv.Sniffer().sniff(csvfile.read(1024), delimiters=";,")
                    except Exception:
                        # Sometimes the sniffer fails, we switch back to the default ... of french statistical data
                        dialect = None
                        delimiter = ";"

                    log.debug(
                        f"dialect.delimiter = {dialect.delimiter if dialect is not None else delimiter}, encoding = {encoding}, confidence = {confidence}"
                        )
                    kwargs['engine'] = "python"
                    if dialect:
                        kwargs['dialect'] = dialect
                    else:
                        kwargs['delimiter'] = delimiter
                    kwargs['encoding'] = encoding
                    data_frame = reader(data_file, **kwargs)

            else:
                data_frame = reader(data_file, **kwargs)

        except Exception as e:
            log.info(f'Error while reading {data_file}')
            raise e

        gc.collect()
        return data_frame

    def save_data_frame_to_hdf5(self, data_frame, **kwargs):
        """Save a data frame in the HDF5 file format."""
        hdf5_file_path = self.survey.hdf5_file_path
        log.info(f"Inserting table {self.name} in HDF file {hdf5_file_path}")
        store_path = self.name
        try:
            data_frame.to_hdf(hdf5_file_path, store_path, append = False, **kwargs)
        except (TypeError, NotImplementedError):
            log.info(f"Type problem(s) when creating {store_path} in {hdf5_file_path}")
            dtypes = data_frame.dtypes
            # Checking for strings
            converted_dtypes = dtypes.isin(['mixed', 'unicode'])
            if converted_dtypes.any():
                log.info(f"The following types are converted to strings \n {dtypes[converted_dtypes]}")
                # Conversion to strings
                for column in dtypes[converted_dtypes].index:
                    data_frame[column] = data_frame[column].copy().astype(str)

            # Checking for remaining categories
            dtypes = data_frame.dtypes
            converted_dtypes = dtypes.isin(['category'])
            if not converted_dtypes.empty:  # With category table format is needed
                log.info(f"The following types are added as category using the table format\n {dtypes[converted_dtypes]}")
                data_frame.to_hdf(hdf5_file_path, store_path, append = False, format = 'table', **kwargs)

        self.variables = list(data_frame.columns)

    def save_data_frame_to_parquet(self, data_frame):
        """Save a data frame in the Parquet file format."""
        parquet_file_path = self.survey.parquet_file_path

        if not os.path.isdir(parquet_file_path):
            log.warn(f"{parquet_file_path} where to store table {self.name} data does not exist: we create the directory")
            os.makedirs(parquet_file_path)
        self.parquet_file = parquet_file_path + "/" + self.name
        data_frame.to_parquet(self.parquet_file)
        self.variables = list(data_frame.columns)

        self.survey.tables[self.name]["parquet_file"] = self.parquet_file
        self.survey.tables[self.name]["variables"] = self.variables


def clean_data_frame(data_frame):
    """Clean a data frame.

    The following steps are executed:
    - drop empty columns
    - replace empty strings with zeros
    - convert string columns to integers
    """
    data_frame.columns = data_frame.columns.str.lower()
    object_column_names = list(data_frame.select_dtypes(include=["object"]).columns)
    log.info(f"The following variables are to be cleaned or left as strings : \n {object_column_names}")
    for column_name in object_column_names:
        if data_frame[column_name].isnull().all():  #
            log.info(f"Drop empty column {column_name}")
            data_frame.drop(column_name, axis = 1, inplace = True)
            continue

        values = [str(value) for value in data_frame[column_name].value_counts().keys()]
        empty_string_present = "" in values
        if empty_string_present:
            values.remove("")
        all_digits = all([value.strip().isdigit() for value in values])
        no_zero = all([value != 0 for value in values])
        if all_digits and no_zero:
            log.info(f"Replacing empty string with zero for variable {column_name}")
            data_frame.replace(
                to_replace = {column_name: {"": 0}},
                inplace = True,
                )
            log.info(f"Converting string variable {column_name} to integer")
            try:
                data_frame[column_name] = data_frame[column_name].astype("int")
            except OverflowError:
                log.info(f'OverflowError when converting {column_name} to int. Keeping as {data_frame[column_name].dtype}')



import pandas as pd
import os
import unittest
from openfisca_survey_manager import openfisca_survey_manager_location
from openfisca_survey_manager.scripts.build_collection import build_survey_collection


class TestWriteParquet(unittest.TestCase):
    def test_write_parquet_one_file_per_entity(self):
        data_dir = os.path.join(
            openfisca_survey_manager_location,
            'openfisca_survey_manager',
            'tests',
            'data_files',
            'test_parquet_collection',
            )
        os.makedirs(data_dir, exist_ok=True)
        df = pd.DataFrame({'household_id': [1, 2, 3, 4], 'rent': [1100, 2200, 3_000, 4_000], 'household_weight': [550, 1500, 700, 200], 'accommodation_size': [50, 100, 150, 200]})
        filepath = os.path.join(data_dir, 'household.parquet')
        df.to_parquet(filepath)
        df = pd.DataFrame({'person_id': [11, 22, 33, 44, 55], 'household_id': [1, 1, 2, 3, 4],
                           'salary': [1300, 20, 3400, 4_000, 5_000], 'person_weight': [500, 50, 1500, 700, 200],
                           'household_role_index': [0, 1, 0, 0, 0]})
        filepath = os.path.join(data_dir, 'person.parquet')
        df.to_parquet(filepath)
        df2 = pd.read_parquet(filepath)
        assert df.equals(df2)

    def test_write_parquet_multiple_files_per_entity(self):
        collection_name = 'test_multiple_parquet_collection'
        data_dir = os.path.join(
            openfisca_survey_manager_location,
            'openfisca_survey_manager',
            'tests',
            'data_files',
            collection_name,
            )
        os.makedirs(os.path.join(data_dir, "person"), exist_ok=True)
        os.makedirs(os.path.join(data_dir, "household"), exist_ok=True)
        # Create a file config.ini in the current directory
        config = os.path.join(
            data_dir,
            "config.ini",
            )
        with open(config, "w") as f:
            f.write(
                f"""
    [collections]
    collections_directory = {data_dir}
    {collection_name} = {data_dir}/{collection_name}.json

    [data]
    output_directory = {data_dir}
    tmp_directory = /tmp
    """
                )
        # Create a file test_parquet_collection.json in the current directory
        json_file = os.path.join(
            data_dir,
            collection_name + ".json",
            )
        with open(json_file, "w") as f:
            f.write(
                """
    {
    "label": "Test parquet collection",
    "name": "collection_name",
    "surveys": {
    }
    }
    """.replace("collection_name", collection_name)
                )
        df = pd.DataFrame({'household_id': [1, 2], 'rent': [1100, 2200], 'household_weight': [550, 1500], 'accommodation_size': [50, 100]})
        filepath = os.path.join(data_dir, 'household', 'household-0.parquet')
        df.to_parquet(filepath)
        df = pd.DataFrame({'household_id': [3, 4], 'rent': [3_000, 4_000], 'household_weight': [700, 200], 'accommodation_size': [150, 200]})
        filepath = os.path.join(data_dir, 'household', 'household-1.parquet')
        df.to_parquet(filepath)
        df = pd.DataFrame({'person_id': [11, 22], 'household_id': [1, 1], 'salary': [1300, 20], 'person_weight': [500, 50],
                           'household_role_index': [0, 1]})
        filepath = os.path.join(data_dir, 'person', 'person-0.parquet')
        df.to_parquet(filepath)
        df = pd.DataFrame({'person_id': [33, 44, 55], 'household_id': [2, 3, 4], 'salary': [3400, 4_000, 5_000], 'person_weight': [1500, 700, 200],
                           'household_role_index': [0, 0, 0]})
        filepath = os.path.join(data_dir, 'person', 'person-1.parquet')
        df.to_parquet(filepath)
        df2 = pd.read_parquet(filepath)
        assert df.equals(df2)
        collection_name = collection_name
        data_directory_path_by_survey_suffix = {
            "2020": os.path.join(data_dir),
            }
        build_survey_collection(
            collection_name=collection_name,
            data_directory_path_by_survey_suffix=data_directory_path_by_survey_suffix,
            replace_metadata=False,
            replace_data=False,
            source_format="parquet",
            config_files_directory=data_dir,
            keep_original_parquet_file=True,
            )



"""
Test the ability to store parquet files in collections, without converting them to HDF5.
"""

import os
import pytest
import pandas as pd
import logging
from unittest import TestCase

from openfisca_core import periods
from openfisca_survey_manager import openfisca_survey_manager_location, default_config_files_directory
from openfisca_survey_manager.survey_collections import SurveyCollection
from openfisca_survey_manager.scripts.build_collection import add_survey_to_collection
from openfisca_survey_manager.scripts.build_collection import build_survey_collection
from openfisca_survey_manager.scenarios.abstract_scenario import AbstractSurveyScenario
from openfisca_survey_manager.tests import tax_benefit_system
from openfisca_survey_manager.surveys import NoMoreDataError

logger = logging.getLogger(__name__)


@pytest.mark.order(after="test_write_parquet.py::TestWriteParquet::test_write_parquet_one_file_per_entity")
class TestParquet(TestCase):
    data_dir = os.path.join(
        openfisca_survey_manager_location,
        "openfisca_survey_manager",
        "tests",
        "data_files",
        )
    collection_name = "test_parquet_collection"
    survey_name = "test_parquet_survey"

    def test_add_survey_to_collection_parquet(self):
        survey_collection = SurveyCollection(name=self.collection_name)
        survey_file_path = os.path.join(self.data_dir, self.collection_name)
        add_survey_to_collection(
            survey_name=self.survey_name,
            survey_collection=survey_collection,
            parquet_files=[survey_file_path],
            )
        ordered_dict = survey_collection.to_json()
        assert self.survey_name in list(ordered_dict["surveys"].keys())

    def test_build_collection(self):
        collection_name = self.collection_name
        json_file = os.path.join(
            self.data_dir,
            collection_name + ".json",
            )
        with open(json_file, "w") as f:
            f.write(
                """
    {
    "label": "Test parquet collection",
    "name": "collection_name",
    "surveys": {
    }
    }
    """.replace("collection_name", collection_name)
                )
        data_directory_path_by_survey_suffix = {
            "2020": os.path.join(self.data_dir, collection_name),
            }
        build_survey_collection(
            collection_name=collection_name,
            data_directory_path_by_survey_suffix=data_directory_path_by_survey_suffix,
            replace_metadata=False,
            replace_data=False,
            source_format="parquet",
            config_files_directory=self.data_dir,
            )

    @pytest.mark.order(after="test_build_collection")
    def test_load_single_parquet_monolithic(self):
        """
        Test the loading all the data from parquet files in memory.
        """
        # Create survey scenario
        survey_scenario = AbstractSurveyScenario()
        survey_name = self.collection_name + "_2020"
        survey_collection = SurveyCollection.load(
            config_files_directory=self.data_dir,
            collection=self.collection_name,
            )
        survey = survey_collection.get_survey(survey_name)
        table = survey.get_values(
            table="household", ignorecase=True
            )
        input_data_frame_by_entity = table
        survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
        survey_scenario.period = 2020
        survey_scenario.used_as_input_variables = ["rent"]
        survey_scenario.collection = self.collection_name
        period = periods.period('2020-01')
        results = {
            'rent': [],
            'housing_tax': [],
            'income_tax': [],
            }
        data = {
            'collection': self.collection_name,
            'survey': survey_name,
            'input_data_table_by_entity_by_period': {
                period: {
                    'household': 'household',
                    'person': 'person',
                    }
                },
            'config_files_directory': self.data_dir
            }
        survey_scenario.init_from_data(data = data)

        simulation = survey_scenario.simulations["baseline"]
        sim_res = simulation.calculate('housing_tax', period.this_year).flatten().tolist()
        results['housing_tax'] += sim_res
        sim_res = simulation.calculate('rent', period).flatten().tolist()
        results['rent'] += sim_res
        sim_res = simulation.calculate('income_tax', period).flatten().tolist()
        results['income_tax'] += sim_res

        logger.debug(f"{results=}")
        assert len(results['rent']) == 4
        assert (results['rent'] == input_data_frame_by_entity['rent']).all()
        assert (results['housing_tax'] == [500.0, 1000.0, 1500.0, 2000.0])
        assert (results['income_tax'] == [195.00001525878906, 3.0, 510.0000305175781, 600.0, 750.0])

    def test_load_multiple_parquet_monolithic(self):
        """
        Test the loading all the data from parquet files in memory.
        """
        collection_name = 'test_multiple_parquet_collection'
        data_dir = os.path.join(self.data_dir, collection_name)
        # Create survey scenario
        survey_scenario = AbstractSurveyScenario()
        survey_name = collection_name + "_2020"
        survey_collection = SurveyCollection.load(
            config_files_directory=data_dir,
            collection=collection_name,
            )
        survey = survey_collection.get_survey(survey_name)
        table = survey.get_values(
            table="household", ignorecase=True
            )
        input_data_frame_by_entity = table
        survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
        survey_scenario.period = 2020
        survey_scenario.used_as_input_variables = ["rent"]
        survey_scenario.collection = collection_name
        period = periods.period('2020-01')
        results = {
            'rent': [],
            'housing_tax': [],
            'income_tax': [],
            }
        data = {
            'collection': collection_name,
            'survey': survey_name,
            'input_data_table_by_entity_by_period': {
                period: {
                    'household': 'household',
                    'person': 'person',
                    }
                },
            'config_files_directory': data_dir
            }
        survey_scenario.init_from_data(data = data)

        simulation = survey_scenario.simulations["baseline"]
        sim_res = simulation.calculate('housing_tax', period.this_year).flatten().tolist()
        results['housing_tax'] += sim_res
        sim_res = simulation.calculate('rent', period).flatten().tolist()
        results['rent'] += sim_res
        sim_res = simulation.calculate('income_tax', period).flatten().tolist()
        results['income_tax'] += sim_res

        logger.debug(f"{results=}")
        assert len(results['rent']) == 4
        assert (results['rent'] == input_data_frame_by_entity['rent']).all()
        assert (results['housing_tax'] == [500.0, 1000.0, 1500.0, 2000.0])
        assert (results['income_tax'] == [195.00001525878906, 3.0, 510.0000305175781, 600.0, 750.0])

    def test_load_parquet_batch(self):
        """
        Test the batch loading of data from parquet files. This allow loading larger than memory datasets.
        """
        df = pd.read_parquet(
            os.path.join(self.data_dir, self.collection_name, "household.parquet")
            )
        assert len(df) == 4
        assert (df.columns == ["household_id", "rent", "household_weight", "accommodation_size"]).all()
        assert df.rent.sum() == 10300

        collection_name = 'test_multiple_parquet_collection'
        data_dir = os.path.join(self.data_dir, collection_name)
        # Create survey scenario
        survey_scenario = AbstractSurveyScenario()
        survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
        survey_scenario.period = 2020
        survey_scenario.used_as_input_variables = ["rent"]
        survey_scenario.collection = collection_name
        period = periods.period('2020-01')
        results = {
            'rent': [],
            'housing_tax': [],
            'income_tax': [],
            }
        batch_size = 2
        batch_index = 0
        while True:
            try:
                data = {
                    'collection': collection_name,
                    'survey': collection_name + '_2020',
                    'input_data_table_by_entity_by_period': {
                        period: {
                            'household': 'household',
                            'person': 'person',
                            'batch_size': batch_size,
                            'batch_index': batch_index,
                            'batch_entity': 'household',
                            'batch_entity_key': 'household_id',
                            'filtered_entity': 'person',
                            'filtered_entity_on_key': 'household_id',
                            }
                        },
                    'config_files_directory': data_dir
                    }
                survey_scenario.init_from_data(data = data)

                simulation = survey_scenario.simulations["baseline"]
                sim_res = simulation.calculate('housing_tax', period.this_year).flatten().tolist()
                results['housing_tax'] += sim_res
                sim_res = simulation.calculate('rent', period).flatten().tolist()
                results['rent'] += sim_res
                sim_res = simulation.calculate('income_tax', period).flatten().tolist()
                results['income_tax'] += sim_res
                logger.debug("XXXXXXXXXXXXXXXXXXXXXx Next batch XXXXXXXXXXXXXXXXXXXXXx")
                batch_index += 1
            except NoMoreDataError:
                logger.debug("No more data")
                break
        logger.debug(f"{results=}")
        assert len(results['rent']) == 4
        # We check the sum as in CI the results are not in the same order
        assert (sum(results['rent']) == df.rent.sum())
        assert (sum(results['housing_tax']) == sum([500.0, 1000.0, 1500.0, 2000.0]))
        self.assertAlmostEqual(sum(results['income_tax']), sum([195.00001525878906, 3.0, 510.0000305175781, 600.0, 750.0]))


if __name__ == "__main__":
    # openfisca_survey_manager_location = Path(__file__).parent.parent
    # os.environ["CI"] = "True"
    logger.debug(openfisca_survey_manager_location)
    logger.debug(f"Default config files directory: {default_config_files_directory}")
    test = TestParquet()
    test.test_build_collection()
    test.test_load_parquet_batch()
    logger.debug("Done")
